import pprint
from cnvrgv2.proxy import HTTP
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.config import error_messages


class DynamicAttributes:
    def __getattr__(self, name):
        """
        Try to get the requested attribute from the attributes object, if it doesnt exists do an API call.
        This function fires when trying to access an attribute that doesnt exists in the object.
        @param name: The requested attribute
        @return: Attribute value/None
        """
        if name in self.__class__.available_attributes.keys():
            val = self._attributes.get(name, None)
            if val is None:
                response = self._proxy.call_api(route=self._route, http_method=HTTP.GET)
                response_attributes = response.attributes
                response_attributes.pop('slug', None)
                # We want to merge the response attributes with the current attributes in order to not override setter
                self._attributes = {**response_attributes, **self._attributes}
                return self._attributes.get(name, None)
            else:
                return val
        else:
            raise AttributeError("type object {} has no attribute {}".format(self.__class__.__name__, name))

    def __setattr__(self, name, value):
        """
        The function checks if an attribute is a dynamic api attribute (from the class.attributes var)
        If the attribute is dynamic, we perform a type check and apply the change.
        @param name: The attribute name we want to set
        @param value: The attribute value
        @return: None
        """
        if name in self.__class__.available_attributes.keys():
            value_type = type(value)
            expected_type = self.__class__.available_attributes[name]
            if value_type != expected_type:
                bad_format_message = error_messages.ARGUMENT_BAD_TYPE.format(expected_type, value_type)
                raise CnvrgArgumentsError({name: bad_format_message})
            self._attributes[name] = value
        else:
            super().__setattr__(name, value)

    def __str__(self):
        hash_representation = {}
        if hasattr(self, 'slug'):
            hash_representation["slug"] = self.slug

        for key in self.__class__.available_attributes.keys():
            hash_representation[key] = getattr(self, key)

        return pprint.pformat(hash_representation)

    def reload(self):
        """
        Performs hard reload for the module attributes
        @return: None
        """
        response = self._proxy.call_api(route=self._route, http_method=HTTP.GET)
        # We want to merge the response attributes with the current attributes in order to not override setter
        self.__setattr__("_attributes", {**self._attributes, **response.attributes})



from cnvrgv2.context import SCOPE, Context
from cnvrgv2.config import routes, error_messages
from cnvrgv2.errors import CnvrgArgumentsError

from cnvrgv2.modules.flows.flow_version.flow_version import FlowVersion
from cnvrgv2.proxy import Proxy
from cnvrgv2.utils.api_list_generator import api_list_generator


class FlowVersionsClient:
    def __init__(self, context=None):
        self._context = Context(context=context)
        scope = self._context.get_scope(SCOPE.FLOW)

        self._proxy = Proxy(context=self._context)
        self._route = routes.FLOW_VERSIONS_BASE.format(scope["organization"], scope["project"], scope["flow"])

    def get(self, slug):
        """
        Retrieves a flows version by the given title
        @param slug: The slug or title of the requested flow version
        @return: FlowVersion object
        """
        if not slug or not isinstance(slug, str):
            raise CnvrgArgumentsError(error_messages.FLOW_VERSION_GET_FAULTY_SLUG)

        return FlowVersion(context=self._context, slug=slug)

    def list(self, sort="-id"):
        """
        List all flow versions of a specific flows
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @return: Generator that yields flows version objects
        """
        return api_list_generator(
            context=self._context,
            route=self._route,
            object=FlowVersion,
            sort=sort
        )



import os
import re
import fnmatch

from cnvrgv2.errors import CnvrgError
from cnvrgv2.utils.validators import validate_directory_name
from cnvrgv2.utils.storage_utils import (
    match_mimetypes,
    path_is_wildcard,
    get_relative_path,
    append_trailing_slash,
    get_files_and_dirs_recursive,
)


def metadata_generator(paths, mime_filters=None, ignore_spec=None, temp_folder=None):
    """
    Duplicates the generator in order to calculate the resulting files metadata like total files and sizes.
    @param paths: [List] List of strings representing paths and wildcards
    @param mime_filters: [List] Array of accepted mime types
    @param ignore_spec: [PathSpec] object, or use build_cnvrgignore_spec
    @param temp_folder: [String] The location of the temporary folder
    @return: [Dict, Generator] The filtered files metadata + the file generator
    """
    counter = multi_path_filter_generator(paths, mime_filters=mime_filters, ignore_spec=ignore_spec)
    generator = multi_path_filter_generator(paths, mime_filters=mime_filters, ignore_spec=ignore_spec)

    if temp_folder is not None and os.path.exists(temp_folder):
        counter = filter_deleted_paths_generator(counter, temp_folder)
        generator = filter_deleted_paths_generator(generator, temp_folder)

    # Count files and their size
    total_size = 0
    total_files = 0
    for path in counter:
        total_files += 1
        if os.path.isfile(path):
            total_size += os.path.getsize(path)

    return {"total_size": total_size, "total_files": total_files}, generator


def filter_deleted_paths_generator(path_generator, temp_folder):
    """
    Receives a path generator and adds a filter on top of it to filter out deleted files.
    @param path_generator: [Generator]
    @param temp_folder: [String] The location of the temporary folder
    @return: [Generator] Returns filtered paths
    """
    # This function should run only if tmp folder exist (sync/download)
    if not os.path.exists(temp_folder):
        raise CnvrgError("Trying to filter paths using a non-existent temp folder")

    # We need to filter out all paths that are under deleted folders
    deleted_folders = get_files_and_dirs_recursive(root_dir=temp_folder, regex=".*deleted/", force=True)

    for path in path_generator:
        deleted_file = "{}/{}.deleted".format(temp_folder, path)

        # Check that the file is not in a deleted folder
        in_deleted_folder = False
        for deleted_folder in deleted_folders:
            matching_path = deleted_folder.replace(".tmp/", "").replace(".deleted", "")
            regex = re.compile("^{}.*".format(matching_path))
            if regex.match(path):
                in_deleted_folder = True

        if os.path.exists(deleted_file) or in_deleted_folder:
            continue

        yield path


def multi_path_filter_generator(paths, mime_filters=None, ignore_spec=None):
    """
    Generates and filters file/folder paths from the given path list
    @param paths: [List] list of string of folder/file/regex
    @param mime_filters: [List] Array of accepted mime types
    @param ignore_spec: [PathSpec] object, or use build_cnvrgignore_spec
    @return: [Generator] Returns filtered paths
    """
    filters = [path for path in paths if path_is_wildcard(path)]
    for path in multi_path_generator(paths=paths):
        # If we want to ignore specific files (for example via .cnvrgignore)
        if ignore_spec and ignore_spec.match_file(path):
            continue

        # Filter out paths using the regex filters
        matches_filter = False if len(filters) > 0 else True
        for regex in filters:
            if fnmatch.fnmatch(path, regex):
                matches_filter = True
        if not matches_filter:
            continue

        # Filter files by mimetype
        if not match_mimetypes(filters=mime_filters, path=path):
            continue

        # Filter special paths
        if path in ["./", ".", "/", "//"]:
            continue

        yield path


def multi_path_generator(paths):
    """
    Generator that yields all file/folder paths from provided path list.
    @param paths: [List] list of string of folder/file/regex
    @return: [Generator] Returns all paths
    """

    # For pattern matching we will yield ALL files from the current working directory
    # Filtering for said files should be performed OUTSIDE of the generator
    folder_patterns_traversed = []

    for target_path in paths:
        is_dir = os.path.isdir(target_path)
        is_file = os.path.isfile(target_path)

        if is_dir:
            validate_directory_name(target_path)
            # For directories we will use os.walk generator to yield files
            yield from folder_path_generator(root_dir=target_path)
        elif is_file:
            # For files we will yield them right away
            yield target_path
        elif path_is_wildcard(target_path):
            # Calculate which folder we should traverse as part of the regex
            if "/" in target_path or ".." in target_path:
                if "*" in target_path:
                    regex_path = target_path.split("*")[0]
                elif "?" in target_path:
                    regex_path = target_path.split("?")[0]
            else:
                regex_path = "."

            # We will traverse only once per folder and leave the filtering to the caller function
            if regex_path in folder_patterns_traversed:
                continue

            folder_patterns_traversed.append(regex_path)
            yield from folder_path_generator(root_dir=regex_path)


def folder_path_generator(root_dir="."):
    """
    Recursively traverse a given directory and get all of the relevant files and folders within
    @param root_dir: [String] The target directory to traverse
    @return: [Generator] returns all files and directories inside a given path
    """

    for root, dirs, files in os.walk(root_dir, topdown=False, followlinks=True):
        for name in files:
            file_path = get_relative_path(os.path.join(root, name))
            yield file_path
        for name in dirs:
            dir_path = get_relative_path(os.path.join(root, name))
            dir_path = append_trailing_slash(dir_path)
            yield dir_path

    # Add the folder itself, if it's an empty one
    if len(os.listdir(root_dir)) == 0:
        dir_path = append_trailing_slash(get_relative_path(root_dir))
        yield dir_path



import os
import re

from cnvrgv2.data.local_files_handler import LocalFilesHandler
from cnvrgv2.proxy import HTTP
from cnvrgv2.utils.storage_utils import create_dir_if_not_exists
from cnvrgv2.utils.url_utils import urljoin


class FileCompare(LocalFilesHandler):
    def __init__(self, data_owner, paths, metadata, override=False, num_workers=40, chunk_size=1000, queue_size=5000):
        """
        Multithreaded file compare for cnvrg datasets - used to compare between files that should be uploaded
        so we can mark conflicts

        @param data_owner: Cnvrg dataset / project object
        @param paths: [Generator] that lists paths to upload
        @param metadata: [Dict] The metadata on the uploaded files (total files, file size..)
        @param override: Boolean stating whether or not we should re-upload even if the file already exists
        @param num_workers: Number of threads to use for concurrent file upload
        @param chunk_size: File meta chunk size to fetch from the server
        @param queue_size: Max number of file meta to prefetch from the server
        """
        super().__init__(
            data_owner,
            paths=paths,
            metadata=metadata,
            override=override,
            num_workers=num_workers,
            chunk_size=chunk_size,
            queue_size=queue_size
        )

    def _file_handler_function(self, local_path, object_path, progress_bar=None):
        """
        Function to mark conflicts
        @param local_path: file location locally
        @param object_path: file location in bucket
        @return: None
        """
        tmp_folder = "{}/.tmp".format(self.data_owner.working_dir)
        downloaded_file = "{}/{}".format(tmp_folder, local_path)
        if os.path.isfile(downloaded_file):
            new_path = re.sub(tmp_folder, self.data_owner.working_dir, downloaded_file)
            create_dir_if_not_exists(new_path)
            os.rename(downloaded_file, "{}.conflict".format(new_path))

    def _file_collector_function(self, files_metadata):
        """
        Function to collect files metadata from server
        @param files_metadata: Local files metadata to sent to the server for getting files to compare
        @return: Array of files metadata
        """
        data = {
            "files": files_metadata,
            "override": self.override,
        }
        response = self.data_owner._proxy.call_api(
            route=urljoin(self.data_owner._route, "commits", self.data_owner.local_commit, "compare_local"),
            http_method=HTTP.POST,
            payload=data
        )

        blob_array = []
        for blob in response.items:
            blob_array.append({"id": blob.id, **blob.attributes})

        return blob_array



import click

from cnvrgv2.cli.utils import messages
from cnvrgv2.cli.utils.decorators import prepare_command


@click.group(name='workspace')
def workspace_group():
    pass


@workspace_group.command()
@click.option('-s', '--slug', default=None, help=messages.WORKSPACE_SLUG)
@prepare_command()
def start_tensorboard(workspace, logger, slug):
    """
      starts tensorboard associated with the current workspace
    """
    if workspace is None:
        logger.log_and_echo(messages.WORKSPACE_DOES_NOT_EXIST, error=True)

    workspace.start_tensorboard()
    logger.log_and_echo('tensorboard started!')


@workspace_group.command()
@click.option('-s', '--slug', default=None, help=messages.WORKSPACE_SLUG)
@prepare_command()
def stop_tensorboard(workspace, logger, slug):
    """
      stops tensorboard associated with the current workspace
    """
    if workspace is None:
        logger.log_and_echo(messages.WORKSPACE_DOES_NOT_EXIST, error=True)

    workspace.stop_tensorboard()
    logger.log_and_echo('tensorboard session stopped')



from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.config import error_messages, routes
from cnvrgv2.utils.api_list_generator import api_list_generator
from cnvrgv2.utils.validators import attributes_validator, validate_gpu
from cnvrgv2.modules.resources.templates.spark_template import SparkTemplate

required_attributes = [
    'title',
    'cpu',
    'memory',
]

default_attributes = {
    "num_executors": 1.0,
    "compute_type": 'spark_driver'
}


class SparkTemplatesClient:
    def __init__(self, context=None):
        self._context = Context(context=context)
        scope = self._context.get_scope(SCOPE.RESOURCE)
        self._proxy = Proxy(context=self._context)
        self._route = routes.TEMPLATES_BASE.format(
            scope["organization"],
            'spark_driver',
            scope['resource']
        )

    def list(self, sort="-id"):
        """
        List all spark driver compute templates in a specific resource
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @raise: HttpError
        @return: Generator that yields spark driver compute template objects
        """
        return api_list_generator(
            context=self._context,
            route=self._route,
            object=SparkTemplate,
            sort=sort,
        )

    def get(self, slug):
        """
        Retrieves a spark driver compute template by the given slug
        @param slug: The slug of the requested spark driver compute template
        @return: Spark driver compute template object
        """
        if not slug or not isinstance(slug, str):
            raise CnvrgArgumentsError(error_messages.TEMPLATE_GET_FAULTY_SLUG)

        return SparkTemplate(context=self._context, slug=slug)

    def create(
        self,
        title,
        cpu,
        memory,
        **kwargs
    ):
        """
        @param title: Name of the new spark compute template
        @param cpu: CPU cores of the new spark compute template
        @param memory: RAM of the new spark compute template
        @param kwargs: A list of optional attributes for creation
            users: Users with accesss to the new spark compute template
            gpu: GPU Cores for the new spark compute template
            gaudi: Gaudi accelerators num for the new spark compute template
            is_private: is the new spark compute template private or public
            num_executors: number of workers for the new spark compute template
            templates_ids: allowed templates for the new spark compute template worker
            mig_device: mig device type for the new spark compute template
        @return: The newly created spark compute template
        """

        formatted_attributes = {
            "title": title,
            "cpu": float(cpu),
            "memory": float(memory),
            **default_attributes,
            **kwargs,
            "gpu": {
                "count": kwargs.get('gpu'),
                "mig_device": kwargs.get('mig_device')
            },
        }

        formatted_attributes.pop('mig_device', None)

        attributes_validator(
            SparkTemplate.available_attributes,
            formatted_attributes,
            required_attributes
        )

        # Specific instance validations
        validate_gpu(formatted_attributes["gpu"])

        response = self._proxy.call_api(
            route=self._route,
            http_method=HTTP.POST,
            payload=JAF.serialize(type='spark_template', attributes=formatted_attributes)
        )

        slug = response.attributes['slug']
        return SparkTemplate(
            context=self._context,
            slug=slug,
            attributes=response.attributes
        )



from cnvrgv2.config import routes
from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.modules.base.dynamic_attributes import DynamicAttributes
from cnvrgv2.utils.validators import attributes_validator


class OrganizationSettings(DynamicAttributes):

    available_attributes = {
        "default_computes": list,
        "install_dependencies": bool,
        "slack_webhook_url": str,
        "debug_time": int,
        "email_on_error": bool,
        "email_on_success": bool,
        "queued_compute_wait_time": int,
        "idle_enabled": bool,
        "idle_time": int,
        "max_duration_workspaces": int,
        "max_duration_experiments": int,
        "max_duration_endpoints": int,
        "max_duration_webapps": int,
        "automatically_clear_cached_commits": int
    }

    def __init__(self, organization):
        self._context = Context(context=organization._context)
        scope = self._context.get_scope(SCOPE.ORGANIZATION)

        self._proxy = Proxy(context=self._context)
        self._route = routes.ORGANIZATION_SETTINGS.format(scope["organization"])

        self._attributes = {}

    def save(self):
        """
        Save the local settings in the current organization
        @return: None
        """
        self.update(**self._attributes)

    def update(self, **kwargs):
        """
        Updates current organization's settings with the given params
        @param kwargs: any param out of the available attributes can be sent
        @return: None
        """

        attributes_validator(
            available_attributes=OrganizationSettings.available_attributes,
            attributes=kwargs,
        )

        response = self._proxy.call_api(
            route=self._route,
            http_method=HTTP.PUT,
            payload=JAF.serialize(type="settings", attributes={**self._attributes, **kwargs})
        )

        self._attributes = response.attributes



from cnvrgv2._version import __version__
from cnvrgv2.config import routes
from cnvrgv2.config.error_messages import PROXY_UNAUTH_ERROR
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.errors import CnvrgError
from cnvrgv2.modules.datasets_client import DatasetsClient
from cnvrgv2.modules.images.images_client import ImagesClient
from cnvrgv2.modules.registries.registries_client import RegistriesClient
from cnvrgv2.modules.libhub.libraries.libraries_client import LibrariesClient
from cnvrgv2.modules.members.members_client import MembersClient
from cnvrgv2.modules.projects_client import ProjectsClient
from cnvrgv2.modules.resources.clusters_client import ClustersClient
from cnvrgv2.modules.ssh.ssh_client import SshClient
from cnvrgv2.modules.users.user import ROLES, User
from cnvrgv2.modules.users.users_client import UsersClient
from cnvrgv2.modules.organization.organization_settings import OrganizationSettings
from cnvrgv2.proxy import HTTP, Proxy


class Cnvrg:
    def __init__(self, domain=None, email=None, password=None, organization=None, token=None):

        self._context = Context(
            domain=domain,
            user=email,
            password=password,
            organization=organization,
            token=token
        )

        self.check_version_compatibility(__version__)

        self._proxy = Proxy(context=self._context)

        self._init_clients()
        self._organization = organization or self._context.organization

    def me(self):
        """
        Retrieves information about the current user
        @return: A dictionary representing the current logged in user
        """
        users_client = UsersClient(self._context.domain, self._context.token, is_capi=self._context.is_capi)
        return users_client.me()

    def is_admin(self):
        """
        @return: Returns if the logged user is an admin in the current organization or not
        """
        user_membership = next(membership for membership in self.me().roles
                               if membership["organization"] == self._organization)
        return user_membership["access_level"] == ROLES.ADMIN

    def get_users(self):
        """
        Shows all the users in the current organization
        Can only be used by admins
        @return: A list with all the users in the organization
        """
        if not self.is_admin():
            raise CnvrgError(PROXY_UNAUTH_ERROR)

        response = self._proxy.call_api(
            route=routes.ORGANIZATION_USERS.format(self._organization),
            http_method=HTTP.GET,
        )
        users = map(lambda x: User(self._context.domain, x.attributes["token"], x.attributes), response.items)
        return list(users)

    def set_organization(self, organization):
        """
        Replaces the organization with the given one
        @param organization: Organization name (String)
        @return: None
        """
        self._context.set_scope(scope=SCOPE.ORGANIZATION, slug=organization)
        self._init_clients()
        self._organization = organization

    def check_version_compatibility(self, version):
        """
        Check sdk version compatibility against the server
        @param version:
        @return:
        """
        # TODO: IMPLEMENT
        pass

    def _init_clients(self):
        """
        Sets up the clients that are exposed to the user.
        @return: None
        """
        try:
            self.projects = ProjectsClient(self)
            self.datasets = DatasetsClient(self)
            self.libraries = LibrariesClient(self)
            self.images = ImagesClient(self)
            self.registries = RegistriesClient(self)
            self.settings = OrganizationSettings(self)
            self.clusters = ClustersClient(self)
            self.ssh = SshClient(self)
            self.members = MembersClient(self)

        except CnvrgError:
            # TODO: How to handle exceptions
            pass



# GLOBAL
NO_FILE = "File not found"
NO_DIR = "Directory not found"

# VALIDATIONS
FAULTY_KEY = 'Faulty key {0}'
FAULTY_VALUE = 'Faulty value {0}'
MISSING_REQUIRED_VALUE = 'Missing required value {0}'

INVALID_URL = 'URL is invalid'
INVALID_EMAIL = 'Email is invalid'
INVALID_CREDENTIALS = "Could not login using the provided credentials, please check your credentials and try again"

# ORGANIZATION
ORGANIZATION_DOESNT_EXIST = "The requested organization does not exist"

# PROXY
PROXY_HTTP_ERROR = "A server error has occurred."
PROXY_UNAUTH_ERROR = "Unauthorized to perform this action."
PROXY_EMPTY_RESPONSE = "Cannot convert empty response to json"
PROXY_NOT_FOUND_ERROR = "The requested resource could not be found."

# USER
USER_EMPTY_TOKEN = "Can't Login using an empty token"
USER_LOGIN_FAILED = "Can't Authenticate using the provided username and password"
USER_NO_ORGANIZATION = "Please join an organization before using the SDK"

# CONTEXT
CONTEXT_CANT_SAVE = "Can't save context without domain username and organization"
CONTEXT_BAD_ARGUMENTS = "Please provide domain, username and password or token together."
CONTEXT_BAD_ENV_VARIABLES = "Can't authenticate using environment variables/config"

CONTEXT_SCOPE_BAD_SCOPE = "Cannot connect to {0}, please provide a slug"
CONTEXT_SCOPE_BAD_DEPENDENCIES = "Cannot connect to {0} since no {1} is present"

# IMAGE
IMAGE_BAD_TAG = "Image tag is invalid"
IMAGE_BAD_NAME = "Image name is invalid"
IMAGE_BAD_LOGO = "Image logo is not supported"
IMAGE_BAD_REGISTRY = "Image registry must be a string"
IMAGE_GET_FAULTY_PARAMS = "Cannot get image. Please provide a slug or image name and tag (all strings)."
IMAGE_CUSTOM_DOCKERFILE_REQUIRED = "Custom image requires a dockerfile"

# REGISTRY
REGISTRY_BAD_URL = "Registry url must be a string"
REGISTRY_BAD_URL_FORMAT = "Bad registry url format (example: good-url.io)"
REGISTRY_BAD_TYPE = "Registry type is not supported"
REGISTRY_BAD_TITLE = "Registry title must be a string"
REGISTRY_BAD_USERNAME = "Registry username must be a string"
REGISTRY_BAD_PASSWORD = "Registry password must be a string"
REGISTRY_USERNAME_AND_PASSWORD = "Please provide both username and password or neither"

# VOLUME
VOLUME_GET_FAULTY_PARAMS = "Cannot get volume. Please provide a slug or volume name."
NOT_A_VOLUME_OBJECT = "The given argument is not a Volume object. please provide a Volume object."

# COMMIT
COMMIT_FAULTY_SHA1 = "Commit sha1 must be a string"
LOCAL_COMMIT_DOESNT_EXIST_ERROR = "Local commit does not exist. Can't save config"

# PROJECT
PROJECT_GET_FAULTY_SLUG = "Cannot get project with empty slug or a non string slug"
PROJECT_CREATE_FAULTY_NAME = "Cannot create project with empty/non-string name"

# LIBRARY
LIBRARY_GET_FAULTY_SLUG = "Cannot get library with empty slug or a non string slug"

# DATASET
DATASET_GET_FAULTY_SLUG = "Cannot get dataset with empty slug or a non string slug"
DATASET_CREATE_FAULTY_NAME = "Cannot create dataset with empty/non-string name"
DATASET_CREATE_FAULTY_CATEGORY = "Cannot create dataset with empty/non-string category"
DATASET_CREATE_FAULTY_CATEGORY_VALUE = "Dataset category must be one of [general, images, audio, video, text, tabular]"
NOT_A_DATASET_LIST_OBJECT = "datasets must be a list of datasets objects"
NOT_A_DATASET_OBJECT = "dataset parameter missing or not a dataset object"
NOT_A_VALID_DIRECTORY_NAME = "{} is invalid directory name, should contain letters, numbers, '_' , '-' "

# QUERY
QUERY_GET_FAULTY_SLUG = "Cannot get query with empty slug or a non string slug"
QUERY_CREATE_FAULTY_NAME = "Cannot create query with empty/non-string name"
QUERY_CREATE_FAULTY_QUERY = "Cannot create query with empty/non-string query"
QUERY_LIST_FAULTY_PARAMS = "Cannot list with {0}"

# ORGANIZATION
ORGANIZATION_GET_FAULTY_SLUG = "Cannot get organization with empty slug or a non string slug"
ORGANIZATION_CREATE_FAULTY_NAME = "Cannot create organization with empty name or a non string name"
ORGANIZATION_NOT_FOUND = "Organization not found"

# MACHINES
MACHINE_GET_FAULTY_SLUG = "Cannot get machine with empty slug or a non string slug"
MACHINE_GPU_VALUES_WITHOUT_GPU = 'Cannot set gpu values to a non gpu enabled machine'

# SPARK DRIVERS
SPARK_DRIVER_GET_FAULTY_SLUG = "Cannot get spark driver with empty or a non string slug"

# CLUSTER
CLUSTER_GET_FAULTY_SLUG = 'Cannot get cluster with empty or a non string slug'
CLUSTER_ONE_TYPE = 'The given arguments: {} and {} cannot be sent together'

# TEMPLATES
TEMPLATE_GET_FAULTY_SLUG = "Cannot get template with empty slug or a non string slug"
TEMPLATE_FAULTY_MIG_VALUE = 'Mig device must be one of the following: [mig-1g.5g, mig-2g.20g, mig-3g.25g]'
TEMPLATE_FAULTY_GPU = 'Faulty gpu or worker gpu count {0}'
TEMPLATE_FAULTY_MIG = 'Faulty mig value {0}'
TEMPLATE_FAULTY_TYPE = 'Kube template type must be one of the following: [regular, mpi, kubernetes, pytorch, ' \
                       'spark_on_kubernetes] '

# CLONE
CONFIG_YAML_NOT_FOUND = "Couldn't find config file. Please clone project/dataset before syncing."

# EXPERIMENT
MISSING_EXPERIMENT_ARGUMENT = "The function sent as command must get experiment as an argument"
EXPERIMENT_CANNOT_BE_A_FUNCTION = "Function as command is supported in local experiments only." \
                                  " Add local=True or send a string command instead"

# ENDPOINT
ENDPOINT_NOT_RUNNING = "Endpoint {0} is not running"
ENDPOINT_UPDATE_TIMEOUT = "Update endpoint {0} timeout"

# WORKFLOW
WORKFLOW_FINAL_STATE = "Workflow reached a final state {0}, different from desired one"

# FLOW
FLOW_GET_FAULTY_SLUG = "Cannot get flows with empty slug or a non string slug"

# FLOW_VERSION
FLOW_VERSION_GET_FAULTY_SLUG = "Cannot get flows version with empty slug or a non string slug"

# CRON
CRON_ARGUMENTS_MISSING = "Invalid cron syntax. Cron string must contain at least five arguments (format: * * * * *)"
CRON_INVALID_ARGUMENT = "Invalid cron syntax. Cron string accepts * and natural numbers. Optional timezone is a string"
CRON_INVALID_TIMEZONE = "Given timezone does not exist"

# ARGUMENTS
ARGUMENT_BAD_TYPE = "Wrong attribute type, expected {0}, got {1}"
SERIES_NOT_A_LIST = "Series must be a list"
BOTH_GROUP_AND_STEP = "Must include step when using group and visa versa"
ONLY_HEATMAP_CAN_HAVE_GROUPS = "Only Heatmaps support groups and steps"
SERIES_NOT_A_HEATMAP = "Heatmap series must be composed of tuples of the same length made of real numbers"
SERIES_NOT_A_VALID_LIST = "Series must only have real numbers"
EMPTY_KAFKA_BROKERS_LIST = "Must be a list containing one broker url or more"
EMPTY_KAFKA_INPUT_TOPICS = "Must be a list containing one topic or more"
EMPTY_ARGUMENT = "{} cannot be empty"
NOT_LIST = "{} is not a list"
INVALID_CRON_ARGUMENT = "{0} must be a number between {1} and {2}"
FILE_TOTAL = "File total is mandatory for azure callback. Please provide a mutable counter - list containing zero [0]"
NOT_SUPPORTED = "{} are not supported"
SECRETS_NOT_SUPPORTED = "{} are not supported, use update_secret or delete_secret methods"

# GIT
CANT_USE_BOTH_SSH_AND_TOKEN = "Can't use both SSH and Token for authentication with git"
INVALID_SSH_KEY_TYPE = "SSH Key Type needs to be plain/file"
CANT_READ_SSH_KEY_FILE = "Can't read SSH key file"

# CLI - CONFIG FILE
CONFIG_FILE_MISSING = "Config file doesn't exist."
LOCAL_CONFIG_MISSING_DATA_OWNER_SLUG = "Cannot preform operation. Couldn't find project or dataset name" \
                                       " in local config file. Try cloning the project or dataset first."


# CLI - CREDENTIALS
CREDENTIALS_MISSING = "Some credentials are missing. Cannot authenticate, try logging in first."

# CLI - PROJECT
DIRECTORY_ALREADY_LINKED = "Directory is already linked to {0}"
NOT_A_GIT_PROJECT = "Project {0} is not a git project"
PROJECT_NOT_EXIST = "Project {0} is not a cnvrg project"

# CLI - MEMBERS
MEMBER_NOT_VALID_ROLE = "{0} is not a valid role, valid roles are: admin, manager, member, reviewer"

# CLI - SSH
SSH_FAILED_TO_START = "Failed to start ssh. Ssh status: {0}"
SSH_FAILED_GET_REQUIRED_PARAMS = "Failed to get required ssh params. Ssh status: {0}"



import threading
import time

from progress.bar import Bar
from cnvrgv2.utils.converters import convert_bytes


def init_progress_bar_for_cli(name, total_files_size):
    """
    Inits a progress bar for cli purposes. Adds the unit to use and a mutex to the progress bar object
    @param name: Name of the progress bar
    @param total_files_size: total size of files to handle in the progress bar
    @return: The progress bar object
    """
    total_files_size, unit = convert_bytes(total_files_size)
    progress_bar = Bar(name, max=total_files_size, suffix='%(index).2f / %(max).2f ' + unit)
    progress_bar.unit = unit
    progress_bar.mutex = threading.Lock()
    progress_bar.last_next_call = None
    progress_bar.throttled_next_interval = 1  # in seconds
    progress_bar.throttled_next = throttled_next(progress_bar)
    return progress_bar


def throttled_next(self):
    def custom_next(progress):
        current_time = time.time()
        if self.last_next_call is None \
           or current_time > self.last_next_call + self.throttled_next_interval\
           or self.index + progress >= self.max:
            self.next(progress)
            self.last_next_call = current_time
        else:
            self.index = self.index + progress

    return custom_next



from cnvrgv2.config import error_messages, routes
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.modules.registries.registry import Registry
from cnvrgv2.modules.registries.utils import RegistryTypes, validate_registry_params
from cnvrgv2.proxy import HTTP, Proxy
from cnvrgv2.utils.api_list_generator import api_list_generator
from cnvrgv2.utils.json_api_format import JAF


class RegistriesClient:
    def __init__(self, organization):
        self._context = Context(context=organization._context)
        scope = self._context.get_scope(SCOPE.ORGANIZATION)

        self._proxy = Proxy(context=self._context)
        self._route = routes.REGISTRIES_BASE.format(scope["organization"])

    def list(self, sort="-id"):
        """
        List all registries in a specific organization
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @raise: HttpError
        @return: Generator that yields registry objects
        """
        return api_list_generator(
            context=self._context,
            route=self._route,
            object=Registry,
            sort=sort
        )

    def create(self, title, url, type=RegistryTypes.OTHER, username=None, password=None):
        """
        Create a new registry
        @param title: [String] The registry name
        @param url: [String] The registry url
        @param type: [String] The registry type
        @param username: [String] The username with which to connect to the registry
        @param password: [String] The password with which to connect to the registry
        @return: [Registry object] The newly created registry
        """

        validate_registry_params(title=title, url=url, username=username, password=password)
        if (password or username) and not all([username, password]):
            raise CnvrgArgumentsError(error_messages.REGISTRY_USERNAME_AND_PASSWORD)
        if not RegistryTypes.validate_type(type):
            raise CnvrgArgumentsError(error_messages.REGISTRY_BAD_TYPE)

        attributes = {
            "url": url,
            "title": title,
            "registry_type": type,
            "username": username,
            "password": password,
        }
        response = self._proxy.call_api(
            route=self._route,
            http_method=HTTP.POST,
            payload=JAF.serialize(type="registry", attributes=attributes)
        )

        slug = response.attributes['slug']
        return Registry(context=self._context, slug=slug)

    def get(self, slug):
        """
        Retrieves a registry by the given slug
        @param slug: [String] The slug of the requested image
        @return: Registry object
        """

        return Registry(context=self._context, slug=slug)

    def delete(self, slug):
        """
        Delete a registry by its slug
        @param slug: [String] Registry slug
        @return: None
        """

        Registry(context=self._context, slug=slug).delete()



from cnvrgv2.config import routes
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.modules.base.dynamic_attributes import DynamicAttributes
from cnvrgv2.modules.libhub.libraries.library_versions_client import LibraryVersionsClient
from cnvrgv2.proxy import Proxy


class Library(DynamicAttributes):
    available_attributes = {
        "public": bool,
        "name": str,
        "slug": str,
        "owner": str,
        "summary": str,
        "created": str,
        "downloads": int,
        "total_size": int,
        "tags": list
    }

    def __init__(self, context=None, slug=None, attributes=None):
        # Init data attributes
        super().__init__()

        self._context = Context(context=context)

        # Set current context scope to current library
        if slug:
            self._context.set_scope(SCOPE.LIBRARY, slug)

        scope = self._context.get_scope(SCOPE.LIBRARY)

        self._proxy = Proxy(context=self._context)
        self._route = routes.LIBRARY_BASE.format(scope["organization"], scope["library"])
        self._attributes = attributes or {}
        self.slug = scope["library"]

        self._init_clients()

    def _init_clients(self):
        self.versions = LibraryVersionsClient(self)



import click
from cnvrgv2.cli.utils import messages
from cnvrgv2.cli.utils.decorators import prepare_command


@click.group(name='endpoint')
def endpoint_group():
    pass


@endpoint_group.command()
@click.option('-s', '--slug', default=None, help=messages.ENDPOINT_SLUG)
@click.option('-n', '--name', prompt=messages.ENDPOINT_PROMPT_METRIC_NAME, help=messages.ENDPOINT_METRIC_NAME)
@click.option('-x', '--x', prompt=messages.ENDPOINT_PROMPT_METRIC_X, help=messages.ENDPOINT_METRIC_X)
@click.option('-y', '--y', prompt=messages.ENDPOINT_PROMPT_METRIC_Y, help=messages.ENDPOINT_METRIC_Y)
@prepare_command()
def log_metric(endpoint, logger, slug, name, y, x):
    """
        logs a metric to the endpoint
    """
    if endpoint is None:
        logger.log_and_echo(messages.Endpoint_DOES_NOT_EXIST, error=True)

    endpoint.log_metric(name=name, y=y, x=x)



import time

from cnvrgv2.config.error_messages import WORKFLOW_FINAL_STATE
from cnvrgv2.errors import CnvrgFinalStateReached


class WorkflowStatuses:
    PENDING = "pending"
    PENDING_APPROVAL = "pending_approval"
    INITIALIZING = "initializing"
    QUEUED = "queued"
    SCHEDULED = "scheduled"
    RECURRING = "recurring"
    RESTARTING = "restarting"
    ONGOING = "ongoing"
    TERMINATING = "terminating"
    SUCCESS = "success"
    ERROR = "error"
    DEBUG = "debug"
    STOPPED = "stopped"

    FINAL_STATES = [SUCCESS, ERROR, DEBUG, STOPPED]


class WorkflowUtils:

    @staticmethod
    def wait_for_statuses(workflow, statuses, poll_interval=10, max_duration=600):
        """
        Polls on a workflow status until in reaches the wanted status or times out
        @param workflow: The workflow to poll on
        @param statuses: List of required statuses.
        @param poll_interval: Time between polls
        @param max_duration: Polling timeout
        @return: The workflow's status at the end of polling
        """
        start_time = time.time()

        while True:
            if workflow.status in statuses:
                break
            if time.time() >= start_time + max_duration:
                raise TimeoutError()
            # Workflow reached end state different from desired. No need to continue
            if workflow.status in WorkflowStatuses.FINAL_STATES:
                raise CnvrgFinalStateReached(WORKFLOW_FINAL_STATE.format(workflow.status))

            time.sleep(poll_interval)
            workflow.reload()

        return workflow.status



import os
import sys
import click
import logging

from logging.handlers import TimedRotatingFileHandler
from cnvrgv2.config import Config, GLOBAL_CNVRG_PATH


class CnvrgLogger:
    DEFAULT_KEEP_DURATION_DAYS = 7
    CLI_LOGGER = "cli-logger"

    def __init__(self, click):
        # TODO: Assuming config file exist. make sure it exist on cli start
        config = Config()
        keep_duration_days = (
            config.keep_duration_days or
            CnvrgLogger.DEFAULT_KEEP_DURATION_DAYS
        )

        self.click = click
        self.logger = logging.getLogger(CnvrgLogger.CLI_LOGGER)
        self.logger.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s \t [%(levelname)s] > %(message)s')

        dirname = os.path.join(GLOBAL_CNVRG_PATH, "logs")

        if not os.path.isdir(dirname):
            os.mkdir(dirname)

        handler = TimedRotatingFileHandler(dirname + "/cnvrg.log",
                                           when="midnight",
                                           backupCount=keep_duration_days)
        handler.setFormatter(formatter)

        self.logger.addHandler(handler)

    def log_and_echo(self, message, error=False):
        """
        logs the message to the log file and prints it to stdout through click echo function
        @param message: The message to log
        @param error: Boolean. Whenever the log is an error. If it is, program will exit with status code 1
        @return: None
        """
        if error:
            self.click.secho(message, err=error, fg='red')
            self.logger.error(message)
            sys.exit(1)
        else:
            self.click.echo(message, err=error)
            self.logger.info(message)

    def __getattr__(self, item):
        return getattr(self.logger, item)

    @staticmethod
    @click.command()
    @click.option('--days',
                  default=7,
                  prompt='Please specify the number of days to keep logs',
                  help='Number of days to keep logs'
                  )
    def set_logs_keep_duration(days):
        """
        Sets the number of days to keep logs as backup
        """
        config = Config()
        config.keep_duration_days = days
        config.save()



from cnvrgv2.modules.workflows.webapp.webapp import WebappType
from cnvrgv2.config.error_messages import NOT_A_DATASET_LIST_OBJECT, FAULTY_VALUE
from cnvrgv2.context import SCOPE
from cnvrgv2.config import routes
from cnvrgv2.errors import CnvrgArgumentsError

from cnvrgv2.modules.base.workflows_base import WorkflowsBase
from cnvrgv2.modules.dataset import Dataset
from cnvrgv2.modules.workflows import Webapp
from cnvrgv2.utils.validators import validate_types_in_list


class WebappsClient(WorkflowsBase):
    def __init__(self, project):
        super().__init__(Webapp, "Webapp", project._context)

        scope = self._context.get_scope(SCOPE.PROJECT)
        self._route = routes.WEBAPPS_BASE.format(scope["organization"], scope["project"])

    def create(self, webapp_type, file_name, title=None, templates=None, datasets=None, *args, **kwargs):
        """
        Create a new webapp
        @param webapp_type: The type of webapp to create (shiny, dash or voila)
        @param file_name: The file for webapp creating
        @param title: Name of the webapp
        @param templates: List of template names to be used
        @param datasets: List of datasets to connect with the webapp.
        @param args: optional arguments
        @param kwargs: Dictionary. Rest of optional attributes for creation
            image: Image object to create webapp with
            queue: Name of the queue to run this job on
        TODO: Add a list of optional attributes
        @return: The newly created webapp object
        """

        if webapp_type not in [WebappType.SHINY, WebappType.DASH, WebappType.VOILA]:
            raise CnvrgArgumentsError(FAULTY_VALUE.format(webapp_type))

        kwargs = {
            "webapp_type": webapp_type,
            "file_name": file_name,
            **kwargs
        }

        if datasets and not validate_types_in_list(datasets, Dataset):
            raise CnvrgArgumentsError(NOT_A_DATASET_LIST_OBJECT)
        elif datasets:
            kwargs["job_datasets"] = [ds.as_request_params() for ds in datasets]

        return super().create(title, templates, *args, **kwargs)



import boto3
import botocore
from boto3.s3.transfer import TransferConfig

from cnvrgv2.data.clients.base_storage_client import BaseStorageClient
from cnvrgv2.utils.retry import retry
from cnvrgv2.utils.storage_utils import create_dir_if_not_exists

config = TransferConfig(max_concurrency=10, use_threads=True)


class MinioStorage(BaseStorageClient):
    def __init__(self, storage_meta):
        super().__init__(storage_meta)

        props = self._decrypt_dict(storage_meta, keys=["sts_a", "sts_s", "sts_st", "bucket", "region", "endpoint"])

        self.s3props = {
            "endpoint_url": props.get("endpoint"),
            "aws_access_key_id": props.get("sts_a"),
            "aws_secret_access_key": props.get("sts_s"),
            "region_name": props.get("region")
        }
        self.bucket = props.get("bucket")
        self.region = props.get("region")
        self.client = self._get_client()

    @retry(log_error=True)
    def upload_single_file(self, local_path, object_path, progress_bar=None):
        try:
            self.client.upload_file(
                local_path,
                self.bucket,
                object_path,
                Config=config,
                Callback=self.progress_callback(progress_bar)
            )
        except Exception as e:
            print(e)

    @retry(log_error=True)
    def download_single_file(self, local_path, object_path, progress_bar=None):
        try:
            create_dir_if_not_exists(local_path)
            if not object_path:
                return

            self.client.download_file(
                self.bucket,
                object_path,
                local_path,
                Config=config,
                Callback=self.progress_callback(progress_bar)
            )
        except Exception as e:
            raise e

    def _get_client(self):
        botocore_config = botocore.config.Config(max_pool_connections=50)
        return boto3.client('s3', config=botocore_config, verify=self.check_certificate, **self.s3props)



from datetime import datetime

from cnvrgv2.config import routes
from cnvrgv2.modules.base.workflow_instance_base import WorkflowInstanceBase
from cnvrgv2.proxy import Proxy
from cnvrgv2.context import Context, SCOPE


class WebappType:
    SHINY = "rshiny"
    DASH = "dash"
    VOILA = "voila"


class Webapp(WorkflowInstanceBase):
    available_attributes = {
        "webapp_type": str,
        "template_ids": list,
        "template_id": int,
        "compute": str,
        "num_of_exps": int,
        "updated_at": datetime,
        "iframe_url": str,
        "is_public": bool,
        "last_opened": datetime,
        "current_step": int,
        "strip_sources": bool,
        "copy_frequency": int,
        "file_name": str,
        "experiments": list,
        **WorkflowInstanceBase.available_attributes
    }

    def __init__(self, context=None, slug=None, attributes=None):
        self._context = Context(context=context)

        # Set current context scope to current project
        if slug:
            self._context.set_scope(SCOPE.WEBAPP, slug)

        scope = self._context.get_scope(SCOPE.WEBAPP)

        self._proxy = Proxy(context=self._context)
        self._route = routes.WEBAPP_BASE.format(scope["organization"], scope["project"], scope["webapp"])
        self._attributes = attributes or {}
        self._type = "Webapp"
        self.slug = scope["webapp"]

    def start(self):
        """
        Override start from workflows_base to remove functionality.
        start() is only relevant for Endpoints & Workspaces
        """
        raise AttributeError("'Webapp' object has no attribute 'start'")

    def sync_remote(self, commit_msg=None):
        raise NotImplementedError()



import warnings

import urllib3

storage_clients_imports = [
    "from cnvrgv2.data.clients.s3.client import S3Storage",
    "from cnvrgv2.data.clients.azure.client import AzureStorage",
    "from cnvrgv2.data.clients.minio.client import MinioStorage",
    "from cnvrgv2.data.clients.gcp.client import GCPStorage"
]

for storage_client_import in storage_clients_imports:
    try:
        exec(storage_client_import)
    except Exception:
        continue

# Handle client-related warnings
# TODO: move this somewhere else
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
try:
    from boto3.exceptions import PythonDeprecationWarning
    from urllib3.exceptions import InsecureRequestWarning
    warnings.filterwarnings("ignore", category=InsecureRequestWarning)  # Sts download
    warnings.filterwarnings("ignore", category=PythonDeprecationWarning)  # Boto3
except Exception:
    pass


STORAGE_TYPE_GCS = "gcs"
STORAGE_TYPE_AZURE = "azure"
STORAGE_TYPE_S3 = "s3"
STORAGE_TYPE_MINIO = "minio"


def storage_client_factory(refresh_function):
    """
    Returns the relevant storage client depending on the storage meta got from the server
    @param refresh_function: Function that fetches dataset storage meta dictionary
    @return: storage client of type <BaseStorageClient>
    """
    storage_meta = refresh_function()

    if storage_meta is None:
        return None

    # TODO: change all storage types to use refresh_function
    if storage_meta.get("storage") == STORAGE_TYPE_S3:
        return S3Storage(refresh_function)
    if storage_meta.get("storage") == STORAGE_TYPE_MINIO:
        return MinioStorage(storage_meta)
    if storage_meta.get("storage") == STORAGE_TYPE_AZURE:
        return AzureStorage(storage_meta)
    if storage_meta.get("storage") == STORAGE_TYPE_GCS:
        return GCPStorage(storage_meta)



import os
import click

from prettytable import PrettyTable

from cnvrgv2.cli.utils import messages
from cnvrgv2.cli.utils.decorators import prepare_command
from cnvrgv2.errors import CnvrgError


@click.group(name='image')
def image_group():
    """
    Manage images in cnvrg
    """
    pass


@image_group.command()
@click.option('-n', '--name', prompt=messages.IMAGE_PROMPT_NAME, help=messages.IMAGE_HELP_NAME)
@click.option('-t', '--tag', prompt=messages.IMAGE_PROMPT_TAG, help=messages.IMAGE_HELP_TAG)
@click.option('-r', '--registry', default='local', help=messages.IMAGE_HELP_REGISTRY)
@click.option('-l', '--logo', default='', help=messages.IMAGE_HELP_LOGO)
@click.option('-c', '--custom', is_flag=True, default=False, help=messages.IMAGE_HELP_CUSTOM)
@click.option('-rd', '--readme', default='', help=messages.IMAGE_HELP_README)
@click.option('-df', '--dockerfile', default='', help=messages.IMAGE_HELP_DOCKERFILE)
@prepare_command()
def create(cnvrg, logger, name, tag, registry, logo, custom, readme, dockerfile):
    """
    Creates a new image
    """

    # Init readme content
    if readme:
        if not os.path.exists(readme):
            logger.log_and_echo(messages.IMAGE_INVALID_README, error=True)
            return
        with open(readme) as f:
            readme_content = f.read()
    else:
        readme_content = ""

    # Init dockerfile content
    if custom:
        if not dockerfile or not os.path.exists(dockerfile):
            logger.log_and_echo(messages.IMAGE_INVALID_DOCKERFILE, error=True)
            return
        with open(dockerfile) as f:
            dockerfile_content = f.read()
    else:
        dockerfile_content = ""

    new_image = cnvrg.images.create(
        tag=tag,
        name=name,
        logo=logo,
        custom=custom,
        readme=readme_content,
        registry=registry,
        dockerfile=dockerfile_content
    )

    logger.log_and_echo(messages.IMAGE_CREATE_SUCCESS.format(new_image.name))


@image_group.command()
@prepare_command()
def list(cnvrg):
    """
    Lists cnvrg images
    """

    table = PrettyTable()
    table.field_names = ["Slug", "Name", "Tag", "Status", "Registry URL", "Created by"]
    for field in table.field_names:
        table.align[field] = "l"

    for image in cnvrg.images.list():
        created_by = image.created_by or "cnvrg"
        table.add_row([image.slug, image.name, image.tag, image.status, image.registry_url, created_by])

    click.echo(table)


@image_group.command()
@click.option('-s', '--slug', help=messages.IMAGE_HELP_SLUG)
@click.option('-n', '--name', help=messages.IMAGE_HELP_NAME)
@click.option('-t', '--tag', help=messages.IMAGE_HELP_TAG)
@prepare_command()
def get(cnvrg, logger, slug, name, tag):
    """
    Retrieves an image data
    """

    try:
        image = cnvrg.images.get(slug=slug, name=name, tag=tag)
    except CnvrgError as error:
        logger.log_and_echo(message=str(error), error=True)
        return

    table = PrettyTable()
    table.field_names = ["Slug", "Name", "Tag", "Status", "Registry URL", "Created by"]
    for field in table.field_names:
        table.align[field] = "l"

    created_by = image.created_by or "cnvrg"
    table.add_row([image.slug, image.name, image.tag, image.status, image.registry_url, created_by])

    click.echo(table)


@image_group.command()
@click.option('-s', '--slug', prompt=messages.IMAGE_PROMPT_SLUG, help=messages.IMAGE_HELP_SLUG)
@click.option('-l', '--logo', default=None, help=messages.IMAGE_HELP_LOGO)
@click.option('-rd', '--readme', default=None, help=messages.IMAGE_HELP_README)
@prepare_command()
def update(cnvrg, logger, slug, logo, readme):
    """
    Updates an image
    """

    # Init readme content
    if readme:
        if not os.path.exists(readme):
            logger.log_and_echo(messages.IMAGE_INVALID_README, error=True)
            return
        with open(readme) as f:
            readme_content = f.read()
    else:
        readme_content = ""

    cnvrg.images.get(slug=slug).update(logo=logo, readme=readme_content)


@image_group.command()
@click.option('-s', '--slug', help=messages.IMAGE_HELP_SLUG)
@click.option('-n', '--name', help=messages.IMAGE_HELP_NAME)
@click.option('-t', '--tag', help=messages.IMAGE_HELP_TAG)
@prepare_command()
def delete(cnvrg, logger, slug, name, tag):
    """
    Deletes an image
    """

    try:
        image = cnvrg.images.get(slug=slug, name=name, tag=tag)
    except CnvrgError as error:
        logger.log_and_echo(message=str(error), error=True)
        return

    image.delete()



from datetime import datetime
from time import time

import botocore
from boto3 import Session
from boto3.s3.transfer import TransferConfig
from botocore.credentials import RefreshableCredentials
from botocore.session import get_session
from dateutil.tz import tzlocal

from cnvrgv2.data.clients.base_storage_client import BaseStorageClient
from cnvrgv2.utils.retry import retry
from cnvrgv2.utils.storage_utils import create_dir_if_not_exists

config = TransferConfig(max_concurrency=10, use_threads=True)


class S3Storage(BaseStorageClient):
    def __init__(self, refresh_function):
        storage_meta = refresh_function()
        super().__init__(storage_meta)

        props = self._decrypt_dict(storage_meta, keys=["bucket", "region"])

        # An integer number to set the TTL for each session. Beyond this session, it will renew the token.
        # 30 minutes by default which is before the default role expiration of 1 hour
        self.session_ttl = 1800
        self.refresh_function = refresh_function
        self.bucket = props.get("bucket")
        self.region = props.get("region")
        self.client = self._get_client()

    @retry(log_error=True)
    def upload_single_file(self, local_path, object_path, progress_bar=None):
        try:
            self.client.upload_file(
                local_path,
                self.bucket,
                object_path,
                Config=config,
                Callback=self.progress_callback(progress_bar)
            )
        except Exception as e:
            print(e)

    @retry(log_error=True)
    def download_single_file(self, local_path, object_path, progress_bar=None):
        try:
            create_dir_if_not_exists(local_path)
            if not object_path:
                return

            self.client.download_file(
                self.bucket,
                object_path,
                local_path,
                Config=config,
                Callback=self.progress_callback(progress_bar)
            )
        except Exception as e:
            raise e

    def _refresh(self):

        props = self._decrypt_dict(self.refresh_function(), keys=["sts_a", "sts_s", "sts_st", "bucket", "region"])
        credentials = {
            "access_key": props.get("sts_a"),
            "secret_key": props.get("sts_s"),
            "token": props.get("sts_st"),
            "expiry_time": datetime.fromtimestamp(time() + self.session_ttl, tz=tzlocal()).isoformat()
        }
        return credentials

    def _get_client(self):
        session_credentials = RefreshableCredentials.create_from_metadata(
            metadata=self._refresh(),
            refresh_using=self._refresh,
            method="sts_assume_role"
        )
        session = get_session()
        session._credentials = session_credentials
        session.set_config_variable("region", self.region)
        autorefresh_session = Session(botocore_session=session)

        botocore_config = botocore.config.Config(max_pool_connections=50)
        return autorefresh_session.client('s3', config=botocore_config, verify=self.check_certificate)



from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.config import error_messages, routes
from cnvrgv2.utils.api_list_generator import api_list_generator
from cnvrgv2.modules.resources.cluster import Cluster
from cnvrgv2.utils.url_utils import urljoin
import yaml
import os

class ClustersClient:
    def __init__(self, organization):
        self._context = Context(context=organization._context)
        scope = self._context.get_scope(SCOPE.ORGANIZATION)

        self._proxy = Proxy(context=self._context)
        self._route = routes.CLUSTERS_BASE.format(scope["organization"])

    def list(self, sort="-id"):
        """
        List all clusters in a specific resource
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @raise: HttpError
        @return: Generator that yields cluster objects
        """
        return api_list_generator(
            context=self._context,
            route=self._route,
            object=Cluster,
            sort=sort,
        )

    def get(self, slug):
        """
        Retrieves a cluster by the given slug
        @param slug: The slug of the requested cluster
        @return: Cluster object
        """
        if not slug or not isinstance(slug, str):
            raise CnvrgArgumentsError(error_messages.CLUSTER_GET_FAULTY_SLUG)

        return Cluster(context=self._context, slug=slug)

    def create(self,
               build_yaml_path=None,
               kube_config_yaml_path=None,
               resource_name=None,
               provider_name=None,
               domain=None,
               **kwargs):
        """
        Create a new Cluster
        @param build_yaml_path: ccp yaml path
        @param kube_config_yaml_path: kube config yaml path
        @param resource_name: the name of the resource
        @param provider_name: the provider name [aws,gke,eks,dell...]
        @param domain: domain url of existing kubernetes_cluster
        @param kwargs: Dictionary. Rest of optional attributes for creation
            for managed ccp(Cnvrg Cluster Provisioner):
                network: istio as default
            for existing kubernetes_cluster/on premise:
                scheduler: cnvrg_scheduler as default
                namespace: cnvrg as default
                https_scheme: false
                persistent_volumes: false
                gaudi_enabled: false
        @return: Cluster Object
        """

        if build_yaml_path and kube_config_yaml_path:
            raise CnvrgArgumentsError(
                error_messages.CLUSTER_ONE_TYPE.format("build_yaml_path", "kube_config_yaml_path"))

        route = urljoin(self._route, routes.CLUSTER_RESOURCE_REQUEST)
        attributes = {}

        if kube_config_yaml_path:
            if not os.path.exists(kube_config_yaml_path):
                raise CnvrgArgumentsError(error_messages.NO_FILE)

            route = self._route
            kube_config_yaml_content = yaml.safe_load(open(kube_config_yaml_path, "r"))
            attributes = {"domain": domain,
                          "kube_config": str(kube_config_yaml_content),
                          "name": resource_name}
        elif build_yaml_path:
            if not os.path.exists(build_yaml_path):
                raise CnvrgArgumentsError(error_messages.NO_FILE)

            build_yaml_content = yaml.safe_load(open(build_yaml_path, "r"))
            attributes = {"yaml": yaml.dump(build_yaml_content),
                          "provider_name": provider_name}
        elif provider_name:  # No kube_config and no build_yaml_path
            attributes = {"provider_name": provider_name,
                          "resource_name": resource_name}

        attributes.update(**kwargs)

        response = self._proxy.call_api(
            route=route,
            http_method=HTTP.POST,
            payload=attributes
        )

        slug = response.attributes['slug']
        return Cluster(context=self._context, slug=slug)

    def delete(self, slug):
        """
        Delete a cluster by the given slug
        @param slug: The slug of the requested cluster
        @return: None
        """
        if not slug or not isinstance(slug, str):
            raise CnvrgArgumentsError(error_messages.CLUSTER_GET_FAULTY_SLUG)

        Cluster(context=self._context, slug=slug).delete()



from cnvrgv2.context import SCOPE, Context
from cnvrgv2.config import routes, error_messages
from cnvrgv2.errors import CnvrgArgumentsError

from cnvrgv2.modules.flows.flow import Flow
from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.utils.api_list_generator import api_list_generator
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.utils.url_utils import urljoin


class FlowsClient:
    def __init__(self, context=None):
        self._context = Context(context=context)
        scope = self._context.get_scope(SCOPE.PROJECT)

        self._proxy = Proxy(context=self._context)
        self._route = routes.FLOWS_BASE.format(scope["organization"], scope["project"])

    def _create_with_yaml(self, yaml_path):
        """
        Creates a flow from a predefined yaml file
        @param yaml_path: path of the yaml file
        @return: Flow object
        """
        create_with_yaml_url = urljoin(self._route, routes.CREATE_WITH_YAML)
        files_list = [("file", yaml_path)]

        response = self._proxy.call_api(
            route=create_with_yaml_url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type="flows", attributes={}),
            files_list=files_list
        )

        slug = response.attributes['slug']
        return Flow(context=self._context, slug=slug)

    def get(self, slug):
        """
        Retrieves a flows by the given slug
        @param slug: The slug of the requested flows
        @return: Flow object
        """
        # TODO: Something is returning when getting a deleted flows. It fucks things up, check in tests
        if not slug or not isinstance(slug, str):
            raise CnvrgArgumentsError(error_messages.FLOW_GET_FAULTY_SLUG)

        return Flow(context=self._context, slug=slug)

    def create(self, yaml_path=None):
        """
        Creates a new flows with the given name
        @param yaml_path: Path to yaml file to create flows with
        @return: Flow object
        """
        # TODO: Missing in server, Add support for custom name
        # if not name or not isinstance(name, str):
        #     raise CnvrgError(error_messages.PROJECT_CREATE_FAULTY_NAME)

        # attributes = {"title": name}
        if yaml_path is not None:
            return self._create_with_yaml(yaml_path)

        response = self._proxy.call_api(
            route=self._route,
            http_method=HTTP.POST,
            payload=JAF.serialize(type="flows", attributes={})
        )

        slug = response.attributes['slug']
        return Flow(context=self._context, slug=slug)

    def list(self, sort="-id"):
        """
        List all flows in a specific project
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @return: Flow object generator
        """

        return api_list_generator(
            context=self._context,
            route=self._route,
            object=Flow,
            sort=sort
        )

    # DEV-10527: Currently disable. Missing full implementation in the server
    # def delete(self, slugs):
    #    """
    #    Deleting multiple flows
    #    @param slugs: List of flows slugs to be deleted
    #    @return: None
    #    """
    #
    #    attributes = {
    #        "slugs": slugs
    #    }
    #
    #    self._proxy.call_api(
    #        route=self._route,
    #        http_method=HTTP.DELETE,
    #        payload=JAF.serialize(type="flows", attributes=attributes)
    #    )



import click
from cnvrgv2.cnvrg import UsersClient
from cnvrgv2.cli.utils import messages
from cnvrgv2.errors import CnvrgHttpError, CnvrgFileError
from cnvrgv2.config import Config


@click.command()
@click.option('-d', '--domain', default=None, help=messages.LOGIN_PROMPT_DOMAIN)
@click.option('-e', '--email', default=None, help=messages.LOGIN_PROMPT_EMAIL)
@click.option('-p', '--password', default=None, help=messages.LOGIN_PROMPT_PASSWORD, hide_input=True)
@click.option('-o', '--organization', default=None, help=messages.LOGIN_ORGANIZATION_HELP)
@click.option('-t', '--auth-token', is_flag=True, default=False, help=messages.LOGIN_HELP_AUTH_TOKEN)
def login(domain, email, password, organization, auth_token):

    config = Config()
    config_domain = None
    user_token = None
    if password is None:
        password = ""

    if domain is None:
        config_domain = config.domain
        domain = config_domain or click.prompt(messages.LOGIN_PROMPT_DOMAIN, type=str)

    # The guideline in this block is to receive both email and pass together from the user or config file.
    if email is None:
        config_email = config.user
        config_token = config.token

        if config_email and config_token and config_domain:
            click.echo(messages.LOGIN_ALREADY_LOGGED_IN)
            return
        else:
            email = click.prompt(messages.LOGIN_PROMPT_EMAIL, type=str)
            if auth_token:
                user_token = click.prompt(messages.LOGIN_PROMPT_AUTH_TOKEN, type=str, hide_input=True)
            else:
                password = click.prompt(messages.LOGIN_PROMPT_PASSWORD, type=str, hide_input=True)
    else:
        if auth_token:
            user_token = click.prompt(messages.LOGIN_PROMPT_AUTH_TOKEN, type=str, hide_input=True)
        else:
            password = password or click.prompt(messages.LOGIN_PROMPT_PASSWORD, type=str, hide_input=True)

    try:
        uc = UsersClient(domain=domain, token=user_token)
        token, default_organization = uc.login(user=email, password=password)

        config.domain = domain
        config.organization = organization or default_organization
        config.user = email
        if user_token is not None:
            config.token = user_token
        else:
            config.token = token
        config.check_certificate = False
        config.save()

        click.echo(messages.LOGIN_SUCCESS.format(email))

    except CnvrgHttpError:
        click.echo(messages.LOGIN_INVALID_CREDENTIALS, err=True)


@click.command()
def logout():
    try:
        config = Config()
        config.remove_config_fields("user", "token")
        click.echo(messages.LOGOUT_SUCCESS)

    except CnvrgFileError:
        click.echo(messages.LOGOUT_CONFIG_MISSING, err=True)



from datetime import datetime
from cnvrgv2.config import routes
from cnvrgv2.config.error_messages import EMPTY_ARGUMENT, NOT_LIST
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.modules.base.data_owner import DataOwner
from cnvrgv2.proxy import HTTP
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.utils.log_utils import timestamp_for_logs, MAX_LOGS_PER_SEND, LOGS_TYPE_OUTPUT
from cnvrgv2.utils.url_utils import urljoin


class TagTypes:
    SINGLE_TAG = "single"
    LINECHART_TAG = "linechart"


class WorkflowInstanceBase(DataOwner):
    available_attributes = {
        "type": str,
        "title": str,
        "slug": str,
        "status": str,
        "username": str,
        "created_at": datetime,
        "start_time": datetime,
        "end_time": datetime,
        "start_commit": dict,
        "commit": str,
        "last_commit": str,
        "output_folder": str,
        "output_dir": str,
        "git_branch": str,
        "git_commit": str,
        "git": dict,
        "local_folders": list,
        "current_status": str,
        "compute_name": str,
        "external_disk": dict,
        "image_slug": str,
        "image_name": str,
        "datasets": list,
        "job_datasets": list
    }

    def start(self):
        """
        Starts the current workflow
        @return: The started workflow object
        """
        start_url = urljoin(self._route, routes.WORKflOW_START_SUFFIX)

        return self._proxy.call_api(
            route=start_url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes={})
        )

    def sync_remote(self, commit_msg=None):
        """
        Syncing the current workflow
        @return: None
        """
        sync_url = urljoin(self._route, routes.WORKFLOW_SYNC_SUFFIX)

        attributes = {}
        if commit_msg:
            attributes["commit_msg"] = commit_msg

        self._proxy.call_api(
            route=sync_url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes=attributes)
        )

    def stop(self, sync=True):
        """
        Stops the current workflow
        @param sync: sync workflow before stop
        @return: None
        """
        stop_url = urljoin(self._route, routes.WORKFLOW_STOP_SUFFIX)

        attributes = {
            "sync": sync
        }

        self._proxy.call_api(
            route=stop_url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes=attributes)
        )

    def restart(self):
        """
        Restarts the current workflow
        @return: None
        """
        restart_url = urljoin(self._route, routes.WORKFLOW_RESTART_SUFFIX)

        self._proxy.call_api(
            route=restart_url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes={})
        )

    def update(self, **kwargs):
        """
        Updates the current workflow
        @param kwargs: Dictionary of attributes to update.
        supported attributes: title, is_public
        @return: The updated workflow
        """
        if not kwargs:
            raise CnvrgArgumentsError(EMPTY_ARGUMENT.format("kwargs"))

        self._proxy.call_api(
            route=self._route,
            http_method=HTTP.PUT,
            payload=JAF.serialize(type=self._type, attributes=kwargs)
        )

        self.reload()
        return self

    def delete(self):
        """
        Deletes the current workflow
        @return: None
        """
        self._proxy.call_api(
            route=self._route,
            http_method=HTTP.DELETE
        )

    def tag(self, tag_type, key, value, *args, **kwargs):
        """
        Add new tag to workflow
        @param tag_type: string, 'single' or 'linechart'
        @param key: key of the tag
        @param value: value of the tag
        @param args:
        @param kwargs: additional arguments:
        xs: list of x values
        ys: list of y values
        x_axis:
        y_axis:
        grouping:
        key_type:
        @return:
        """
        tag_url = urljoin(self._route, routes.WORKFLOW_TAG_SUFFIX)
        attributes = {
            "tag_type": tag_type,
            "key": key,
            "value": value,
            **kwargs
        }

        return self._proxy.call_api(
            route=tag_url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes=attributes)
        )

    def write_logs(self, logs, log_type=LOGS_TYPE_OUTPUT):
        """
        write logs for the current workflow
        @param logs: list of logs to write
        @param log_type: log level
        @return: None
        """
        if logs is None:
            return

        write_logs_route = urljoin(self._route, routes.WORKFLOW_WRITE_LOGS_SUFFIX)

        if isinstance(logs, str):
            logs = [logs]

        for i in range(0, len(logs), MAX_LOGS_PER_SEND):
            attributes = {
                "logs": logs[i:i + MAX_LOGS_PER_SEND],
                "log_level": log_type,
                "timestamp": timestamp_for_logs()
            }

            self._proxy.call_api(
                route=write_logs_route,
                http_method=HTTP.POST,
                payload=JAF.serialize(type=self._type, attributes=attributes)
            )

    def get_logs(self, filter=None, **kwargs):
        """
        Get logs of the current workflow
        @param filter: list of logs type to filter ('stdout', 'cnvrg-info', 'cnvrg-error')
        @return: List of the logs
        """
        # TODO: Check that it works after server changes
        #  (currently results are not in JAF format, filter expects string with " and not ', so current doesn't work)
        if filter is None:
            filter = ["all"]
        elif type(filter) is not list:
            raise CnvrgArgumentsError(NOT_LIST.format("filter"))

        get_logs_url = urljoin(self._route, routes.WORKFLOW_GET_LOGS_SUFFIX)

        attributes = {
            "filter": str(filter),
            "search": "",
            **kwargs
        }

        return self._proxy.call_api(
            route=get_logs_url,
            http_method=HTTP.GET,
            payload=attributes
        )

    def _validate_config_ownership(self):
        return False

    def sync(self, job_slug=None, git_diff=False, progress_bar_enabled=False):
        return NotImplemented

    def save_config(self):
        return NotImplemented

    def download(self, sync=False, progress_bar_enabled=False, commit_sha1=None):
        return NotImplemented

    def upload(self, sync=False, job_slug=None, progress_bar_enabled=False, git_diff=False):
        return NotImplemented

    def clone(self, progress_bar_enabled=False, override=False):
        return NotImplemented



import json
import os

from cnvrgv2.data.remote_files_handler import RemoteFilesHandler
from cnvrgv2.proxy import HTTP
from cnvrgv2.utils.url_utils import urljoin


class ArtifactsDownloader(RemoteFilesHandler):
    def __init__(
        self,
        data_owner,
        num_workers=40,
        queue_size=5000,
        chunk_size=1000,
        force=False,
        base_commit_sha1=None,
        commit_sha1=None,
        progress_bar_enabled=False,
    ):
        """
        Multithreaded file downloader - download artifacts files from server (by compare commit to base_commit)
        @param data_owner: Cnvrg dataset / project object
        @param num_workers: Number of threads to handle files
        @param queue_size: Max number of file meta to put in queue
        @param chunk_size: File meta chunk size to fetch from the server
        @param force: Force rewrite existing files
        @param base_commit_sha1: Base commit sha1 of the comparision
        @param commit_sha1: Commit sha1 of the comparision
        @param progress_bar_enabled: Boolean indicating whenever or not to print a progress bar. In use of the cli
        """
        self.mode = "files"
        self.compare_url = urljoin(data_owner._route, "commits", commit_sha1,
                                   "compare") + "?page[after]={}&page[size]={}&sort=id&mode={}"

        self.compare_request_data = {
            "base_commit_sha1": base_commit_sha1,
            "filter": json.dumps({
                "operator": 'OR',
                "conditions": [
                    {
                        "key": 'fullpath',
                        "operator": 'like',
                        "value": "*",
                    }
                ],
            })
        }

        super().__init__(
            data_owner,
            num_workers=num_workers,
            queue_size=queue_size,
            chunk_size=chunk_size,
            force=force,
            base_commit_sha1=base_commit_sha1,
            commit_sha1=commit_sha1,
            progress_bar_enabled=progress_bar_enabled
        )

        total_files = data_owner._proxy.call_api(
            route=self.compare_url.format('', 0, "files"),
            http_method=HTTP.GET,
            payload=self.compare_request_data
        ).meta["total"]

        total_folders = data_owner._proxy.call_api(
            route=self.compare_url.format('', 0, "folders"),
            http_method=HTTP.GET,
            payload=self.compare_request_data
        ).meta["total"]

        self.total_files = total_files + total_folders

    def _collector_function(self, page_after=None):
        """
        Function to collect files that should be downloaded
        @param page_after: The id of the next file that the iteration of the pagination should start from
        @return: Should return array of files metadata
        """

        response = self.data_owner._proxy.call_api(
            route=self.compare_url.format(page_after, 1000, self.mode),
            http_method=HTTP.GET,
            payload=self.compare_request_data
        )

        # Once we are out of files we want to move to downloading (empty) folders
        if response.meta["next"] == '' and self.mode == "files":
            self.mode = "folders"
            return {
                "file_dict": [],
                "total_files": self.total_files,
                "total_files_size": 0,
                "next": -100  # placeholder in order to nullify the cursor
            }

        file_dict = []
        for file in response.items:
            file_dict.append(dict(file.attributes))

        return {
            "file_dict": file_dict,
            "total_files": self.total_files,
            "total_files_size": response.meta["total_files_size"],
            "next": response.meta["next"]
        }

    def _handle_file_function(self, local_path, progress_bar=None, **kwargs):
        """
        Function that download single file
        @param local_path: File location locally
        @param progress_bar: A progress bar object to be used during the download
        @param kwargs: Needs to be object_path of the file in the bucket
        @return: None
        """
        if self.force or not os.path.exists(local_path):
            self.storage_client.download_single_file(local_path, kwargs["object_path"], progress_bar)

        self.handle_queue.task_done()
        self.progress_queue.put(local_path)



class StorageUnits:
    B = "B"
    KB = "KB"
    MB = "MB"
    GB = "GB"
    TB = "TB"


def convert_bytes(bytes_to_convert, unit=None):
    """
    converts the given bytes to the given unit, or to the biggest non fraction unit.
    @param bytes_to_convert: bytes value to convert
    @param unit: unit to convert to, from STORAGE_UNITS Enum
    @return: The converted value and the unit
    """
    B = float(bytes_to_convert)
    KB = float(1024)
    MB = float(KB ** 2)  # 1,048,576
    GB = float(KB ** 3)  # 1,073,741,824
    TB = float(KB ** 4)  # 1,099,511,627,776

    units_table = {
        "B": B,
        "KB": KB,
        "MB": MB,
        "GB": GB,
        "TB": TB
    }

    if unit is not None:
        if unit == "B" or B == 0:
            return B, unit
        return B/units_table[unit], unit

    converted_result = None
    if B < KB:
        converted_result = B, StorageUnits.B
    elif KB <= B < MB:
        converted_result = B/KB, StorageUnits.KB
    elif MB <= B < GB:
        converted_result = B/MB, StorageUnits.MB
    elif GB <= B < TB:
        converted_result = B/GB, StorageUnits.GB
    elif TB <= B:
        converted_result = B/TB, StorageUnits.TB

    return converted_result



from cnvrgv2.config import routes
from cnvrgv2.utils.validators import validate_user_role
from cnvrgv2.modules.base.dynamic_attributes import DynamicAttributes
from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.config.error_messages import FAULTY_VALUE


class ROLES:
    ADMIN = 'admin'
    MANAGER = 'manager'
    MEMBER = 'member'
    REVIEWER = 'reviewer'
    ALL_ROLES = [ADMIN, MANAGER, MEMBER, REVIEWER]


class Member(DynamicAttributes):
    available_attributes = {
        "username": str,
        "role": str,
        "access_level": str,
        "email": str,
    }

    def __init__(self, context=None, attributes=None, email=None):
        self._context = Context(context=context)

        if attributes:
            self._email = attributes["email"]

        if email:
            self._email = email

        if self._email:
            self._context.set_scope(SCOPE.MEMBER, self._email)

        scope = self._context.get_scope(SCOPE.MEMBER)

        self._proxy = Proxy(context=self._context)
        self._route = routes.MEMBER_BASE.format(scope["organization"], scope["member"])
        self._attributes = attributes or {}

    @property
    def email(self):
        return self._email

    def revoke(self):
        """
        Delete the membership profile
        """

        self._proxy.call_api(route=self._route, http_method=HTTP.DELETE)

    def update(self, role):
        """
        Update the membership profile
        """

        if not validate_user_role(role):
            raise CnvrgArgumentsError(FAULTY_VALUE.format(role))

        self._proxy.call_api(
            route=self._route,
            http_method=HTTP.PUT,
            payload=JAF.serialize(type="users", attributes={"role": role})
        )




import time
from datetime import datetime

from cnvrgv2.config import routes
from cnvrgv2.config.error_messages import ARGUMENT_BAD_TYPE, EMPTY_ARGUMENT, NOT_LIST
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.data import ArtifactsDownloader
from cnvrgv2.errors import CnvrgArgumentsError, CnvrgError
from cnvrgv2.modules.base.workflow_instance_base import TagTypes, WorkflowInstanceBase
from cnvrgv2.modules.workflows.workflow_utils import WorkflowStatuses, WorkflowUtils
from cnvrgv2.proxy import HTTP, Proxy
from cnvrgv2.utils.chart_utils import BarChart, Chart, Heatmap, LineChart, ScatterPlot
from cnvrgv2.utils.env_helper import ENV_KEYS
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.utils.log_utils import (
    LOGS_TYPE_ERROR, LOGS_TYPE_INFO, LOGS_TYPE_OUTPUT, LOGS_TYPE_WARNING,
    timestamp_for_logs
)
from cnvrgv2.utils.url_utils import urljoin
from cnvrgv2.utils.api_list_generator import api_list_generator
from cnvrgv2.modules.file import File


class Experiment(WorkflowInstanceBase):
    available_attributes = {
        "input": str,
        "href": str,
        "full_href": str,
        "remote": str,
        "ma": str,
        "tags": dict,
        "is_running": bool,
        "terminal_url": str,
        "termination_time": datetime,
        **WorkflowInstanceBase.available_attributes
    }

    CHART_CLASSES = {
        'none': LineChart,
        'scatter': ScatterPlot,
        'bar': BarChart,
        'heatmap': Heatmap
    }

    def __init__(self, context=None, slug=None, attributes=None):
        super().__init__()
        self._context = Context(context=context)

        # Set current context scope to current project
        if slug:
            self._context.set_scope(SCOPE.EXPERIMENT, slug)

        self.scope = self._context.get_scope(SCOPE.EXPERIMENT)

        self._proxy = Proxy(context=self._context)
        self._route = routes.EXPERIMENT_BASE.format(
            self.scope["organization"],
            self.scope["project"],
            self.scope["experiment"]
        )
        self._attributes = attributes or {}
        self._type = "Experiment"
        self.slug = self.scope["experiment"]

    def restart(self):
        """
        restarts an experiment
        @return: The restarted experiment object
        """
        return super().start()

    def finish(self, exit_status):
        # TODO: remove after uniting finish and stop in server.
        #  Ask Leah if experiment should have finish as syntax (product wise)
        """
        Finishes the current experiment
        @param exit_status: exit status of the experiment
        @return: The finished experiment
        """
        finish_url = urljoin(self._route, routes.EXPERIMENT_FINISH_SUFFIX)
        attributes = {
            "exit_status": exit_status
        }

        return self._proxy.call_api(
            route=finish_url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes=attributes)
        )

    def start(self):
        """
        Override start from workflows_base to remove functionality.
        start() is only relevant for Endpoints & Workspaces
        """
        raise AttributeError("'Experiment' object has no attribute 'start'")

    def restart(self):
        """
        Override restart from workflows_base to remove functionality.
        restart() is only relevant for webapps
        """
        raise AttributeError("'Experiment' object has no attribute 'restart'")

    def log_metric(self, chart: Chart):
        """
        Logs a chart of metrics for the current experiment
        @param chart: A chart object
        @return: The created chart object
        """

        create_chart_url = urljoin(self._route, routes.EXPERIMENT_CHARTS_SUFFIX)

        response = self._proxy.call_api(
            route=create_chart_url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes=chart.to_dict())
        )

        NewChartClass = self.CHART_CLASSES.get(chart.chart_type)

        chart_kwargs = {**response.attributes, **response.attributes.get('settings')}
        chart_kwargs.pop('chart_type', None)

        new_chart = NewChartClass(**chart_kwargs)

        return new_chart

    def as_env(self):
        """
        @return: A dict representing current experiment for env use
        """
        return {
            ENV_KEYS["current_job_id"]: self.slug,
            ENV_KEYS["current_job_type"]: self._type,
            ENV_KEYS["current_project"]: self.scope["project"],
            ENV_KEYS["current_organization"]: self.scope["organization"],
        }

    def pull_artifacts(self, wait_until_success=False, poll_interval=10, commit_sha1=None):
        """
        pulls current experiment's artifacts to the local working dir
        @param wait_until_success: Wait until current experiment is done before pulling artifacts
        @param poll_interval: Is wait_until_success is True, time between status poll loops
        @return: None
        """
        if wait_until_success:
            WorkflowUtils.wait_for_statuses(self, WorkflowStatuses.SUCCESS, poll_interval=poll_interval)

        self.reload()

        commit = commit_sha1 if commit_sha1 is not None else self.last_commit
        # Start commit for git project is null
        base_commit = self.start_commit["sha1"] if self.start_commit else None
        downloader = ArtifactsDownloader(self, base_commit_sha1=base_commit, commit_sha1=commit)
        while downloader.in_progress:
            time.sleep(1)

        if downloader.errors:
            raise CnvrgError(downloader.errors.args)
        return downloader

    def log_param(self, key, value=None):
        """
        Adds the given tag to the current experiment
        @param key: [String] The tag name
        @param value: [String] The tag value
        @return: None
        """
        # TODO: add run async to this function and rest of log functions
        return self.tag(tag_type=TagTypes.SINGLE_TAG, key=key, value=value)

    def log_images(self, file_paths):
        """
        Saves the given images to the current experiment as a new commit
        Note that only images that will be saved via this function
        Will show as experiment visuals
        @param file_paths: list of paths of artifacts to save
        @return: None
        """
        commit_msg = "Log Images Commit"
        self.put_files(file_paths, message=commit_msg, job_slug=self.slug, tag_images=True)

    def log_artifacts(self, paths=None, git_diff=False, work_dir=None):
        """
        Saves the given artifacts to the current experiment as a new commit
        @param paths: list of paths of artifacts to save
        @param git_diff: upload files from git diff output in addition to the given paths
        @param work_dir: working directory param, path to files. i.e running command from dir A, files are in dir A/B.
        instead of sending paths B/*.*, set working_dir to B, and send path as *.*
        @return: None
        """
        if not paths and not git_diff:
            raise CnvrgArgumentsError(EMPTY_ARGUMENT.format("file_paths"))
        if not isinstance(paths, list):
            raise CnvrgArgumentsError(NOT_LIST.format("paths"))

        if work_dir:
            self.working_dir = work_dir

        commit_msg = "Log Artifacts Commit"
        self.put_files(paths, message=commit_msg, job_slug=self.slug, git_diff=git_diff)
        self.reload()

    def get_utilization(self):
        """
        Get experiment's utilization stats
        @return: Experiment's system stats data
        """
        return self._proxy.call_api(
            route=urljoin(self._route, routes.EXPERIMENT_GET_UTILIZATION_SUFFIX),
            http_method=HTTP.GET
        )

    def log(self, logs, log_level=LOGS_TYPE_INFO, timestamp=None):
        """
        Method to add logs to the experiment log
        @param logs: an array of logs you want to send
        @param log_level: level of the logs, exists in log_utils
        @param timestamp: timestamp for the logs, UTC now by default
        """
        if timestamp is None:
            timestamp = timestamp_for_logs()

        if type(timestamp) is not str:
            raise CnvrgArgumentsError({"timestamp": ARGUMENT_BAD_TYPE % ("str", type(timestamp))})

        if log_level not in (LOGS_TYPE_OUTPUT, LOGS_TYPE_ERROR, LOGS_TYPE_INFO, LOGS_TYPE_WARNING):
            raise CnvrgArgumentsError({"level": ARGUMENT_BAD_TYPE % ("logs level enum", log_level)})

        if type(logs) is not list:
            if type(logs) is str:
                # give the user the option to only send one string
                logs = [logs]
            else:
                raise CnvrgArgumentsError({"level": ARGUMENT_BAD_TYPE % ("str", type(logs))})

        return self._proxy.call_api(
            route=urljoin(self._route, routes.EXPERIMENT_WRITE_LOGS),
            http_method=HTTP.POST,
            payload={
                "logs": logs,
                "log_level": log_level,
                "timestamp": timestamp
            }
        )

    def logs(self):
        """
        Retrieval of the last 40 logs of the experiment
        @return: a list of the last 40 logs
        """
        response = self._proxy.call_api(
            route=urljoin(self._route, routes.EXPERIMENT_LOGS_SUFFIX),
            http_method=HTTP.GET,
        )

        return response.attributes["logs"]

    def _rerun(self, sync=True, prerun=False, requirements=False):
        """
        Rerun experiment that's currently in debug mode
        @param sync: sync before rerunning
        @param prerun: run prerun.sh script
        @param requirements: install requirements file
        @return: None
        """
        rerun_url = urljoin(self._route, routes.EXPERIMENT_RERUN_SUFFIX)
        attributes = {
            "sync": sync,
            "prerun": prerun,
            "requirements": requirements
        }

        self._proxy.call_api(
            route=rerun_url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes=attributes)
        )

    def list_charts(self):
        """
        List charts of an experiment
        @return: List of charts for the experiment
        """
        charts_url = urljoin(self._route, routes.EXPERIMENT_CHARTS_SUFFIX)

        response = self._proxy.call_api(
            route=charts_url,
            http_method=HTTP.GET
        )

        charts_list = []
        for chart_jaf in response.items:
            chart_type = chart_jaf.attributes.get('settings').get('chart_type')
            ChartClass = self.CHART_CLASSES.get(chart_type)

            chart_kwargs = {**chart_jaf.attributes, **chart_jaf.attributes.get('settings')}
            chart_kwargs.pop('chart_type', None)

            chart = ChartClass(**chart_kwargs)
            charts_list.append(chart)

        return charts_list

    def get_chart(self, key):
        """
        Get charts of an experiment
        @param key: key of chart we wish to get
        @return: Chart with given key
        """
        chart_url = urljoin(self._route, routes.EXPERIMENT_CHART_SUFFIX.format(key))

        response = self._proxy.call_api(
            route=chart_url,
            http_method=HTTP.GET
        )

        # If we couldn't find a chart with given key
        if not len(response.attributes):
            return None

        chart_type = response.attributes.get('settings').get('chart_type')
        ChartClass = self.CHART_CLASSES.get(chart_type)

        chart_kwargs = {**response.attributes, **response.attributes.get('settings')}
        chart_kwargs.pop('chart_type', None)

        chart = ChartClass(**chart_kwargs)

        return chart

    def update_chart(self, key, data, series_name=None):
        """
        Add data to an existing series in a chart
        @param key: key of chart we wish to update
        @param data: new data to add to series
        @param series_name: name of series to update (optional, will update default series without name)
        @return: Chart with given key
        """

        current_chart = self.get_chart(key)

        # Do nothing if we couldn't find given chart
        if not current_chart:
            return None

        # This will validate our data is in the right format for our chart
        formatted_data = current_chart.validate_series(data)

        update_chart_url = urljoin(self._route, routes.EXPERIMENT_CHARTS_SUFFIX)

        attributes = {
            "key": key,
            "series": [{"data": formatted_data, "name": series_name}]
        }

        response = self._proxy.call_api(
            route=update_chart_url,
            http_method=HTTP.PUT,
            payload=JAF.serialize(type=self._type, attributes=attributes)
        )

        ChartClass = self.CHART_CLASSES.get(current_chart.chart_type)

        chart_kwargs = {**response.attributes, **response.attributes.get('settings')}
        chart_kwargs.pop('chart_type', None)

        chart = ChartClass(**chart_kwargs)

        return chart

    def _wait_until_experiment_finish(self, poll_interval=10):
        """
        Busy waits until current experiment run is done
        @param poll_interval: time between status poll loops
        @return: The status of the experiment when done.
        """
        while self.is_running:
            time.sleep(poll_interval)
            self.reload()
        return self.status

    def merge_to_master(self, commit_sha1=None):
        """
        Merge commit to master (used to merge experiment artifacts to master)
        @param commit_sha1: default is latest
        @return: The new commit sha1
        """

        # Importing here to avoid circular dependencies (experiment <-> project)
        from cnvrgv2 import Project
        project = Project(context=self._context, slug=self.scope["project"])
        if project.git:
            return None
        if not commit_sha1:
            commit_sha1 = self.last_commit
        merge_to_master_route = routes.PROJECT_COMMIT_BASE.format(self.scope["organization"], self.scope["project"])
        response = self._proxy.call_api(
            route=urljoin(merge_to_master_route, commit_sha1, "merge_to_master"),
            http_method=HTTP.POST,
            payload={'sha1': commit_sha1}
        )
        return response.attributes['sha1']

    def start_tensorboard(self):
        """
        Starts a tensorboard server in the current experiment
        """
        url = urljoin(self._route, routes.WORKflOW_START_TENSORBOARD_SUFFIX)

        self._proxy.call_api(
            route=url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes={})
        )

    def stop_tensorboard(self):
        """
        Stop the tensorboard session in the current experiment
        """
        url = urljoin(self._route, routes.WORKflOW_STOP_TENSORBOARD_SUFFIX)

        self._proxy.call_api(
            route=url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes={})
        )

    def list_artifacts(self, commit_sha1=None):
        """
        List artifacts of an experiment
        @return: List of artifacts for the experiment
        """
        artifacts_url = urljoin(self._route, routes.WORKFLOW_ARTIFACTS_SUFFIX)
        list_object = File
        response = api_list_generator(
            context=self._context,
            route=artifacts_url,
            object=list_object,
            identifier="fullpath",
            pagination_type="cursor",
            data={
                "sha1": commit_sha1 or self.last_commit
            }
        )

        return response



import abc
import json
import os
import re
import shutil
import time
from datetime import datetime
from itertools import chain

from cnvrgv2.cli.utils.progress_bar_utils import init_progress_bar_for_cli
from cnvrgv2.config import Config
from cnvrgv2.config import CONFIG_FOLDER_NAME
from cnvrgv2.config import error_messages
from cnvrgv2.config.routes import COMMIT_REMOVE_FILES
from cnvrgv2.data import ArtifactsDownloader
from cnvrgv2.data import FileCompare, LocalFileDeleter
from cnvrgv2.data import FileDownloader
from cnvrgv2.data import FileUploader
from cnvrgv2.data import RemoteFileDeleter
from cnvrgv2.errors import CnvrgArgumentsError, CnvrgError, CnvrgFileError
from cnvrgv2.modules.base.dynamic_attributes import DynamicAttributes
from cnvrgv2.modules.file import File
from cnvrgv2.modules.folder import Folder
from cnvrgv2.proxy import HTTP
from cnvrgv2.utils.api_list_generator import api_list_generator
from cnvrgv2.utils.path_generators import metadata_generator
from cnvrgv2.utils.converters import convert_bytes
from cnvrgv2.utils.filter_utils import list_to_multiple_conditions, Operators
from cnvrgv2.utils.storage_utils import (
    cnvrgignore_exists,
    create_cnvrgignore,
    build_cnvrgignore_spec,
    create_dir_if_not_exists,
    get_files_and_dirs_recursive
)
from cnvrgv2.utils.url_utils import urljoin


class DataOwner(DynamicAttributes):
    TMP_FOLDER_NAME = ".tmp"

    def __init__(self):
        # Data attributes
        self._config = Config()
        self._working_dir = None
        self._local_commit = None
        self.query = None

    def storage_meta_refresh_function(self):
        """
        Creates a function that returns an object containing credentials for the storage client
        @return: function for getting credentials for storage client
        """
        # Calculate minutes diff from last update. If over 5 minutes - update storage meta.
        storage_meta_update_time = None
        storage_meta = None
        storage_client_url = urljoin(self._route, "storage_client")
        proxy = self._proxy

        def refresher():
            nonlocal storage_meta_update_time
            nonlocal storage_meta
            minutes_diff = 0

            if storage_meta_update_time is not None:
                minutes_diff = (datetime.now() - storage_meta_update_time).total_seconds() / 60

            if storage_meta is None or minutes_diff > 5:
                response = proxy.call_api(
                    route=storage_client_url,
                    http_method=HTTP.GET
                )
                storage_meta = response.meta["storage"]
                storage_meta_update_time = datetime.now()

            return storage_meta

        return refresher

    @abc.abstractmethod
    def _validate_config_ownership(self):
        """
        Returns if local config file is matched with current object
        """
        pass

    @property
    def local_commit(self):
        """
        Returns the current commit for the data owner
        @return: string denoting current/active commit
        """
        config_commit = None

        if self._validate_config_ownership():
            config_commit = self._config.commit_sha1

        return self._local_commit or config_commit

    @local_commit.setter
    def local_commit(self, commit_sha1):
        self._local_commit = commit_sha1

    @property
    def working_dir(self):
        """
        Returns the local working dir for the data owner
        @return: string denoting the path to the working dir
        """
        if self._working_dir:
            return self._working_dir
        else:
            return os.curdir

    @working_dir.setter
    def working_dir(self, working_dir):
        """
        Sets the working directory
        @param working_dir: string denoting the path to the working dir
        @return: None
        """
        self._working_dir = working_dir

    def _start_commit(self, message="", source="sdk", parent_sha1=None, blank=False, force=False, job_slug=None):
        """
        Starts a new commit
        @param message: Commit message string (Optional)
        @param source: Source of the commit string (Optional)
        @param parent_sha1: Commit sha1 of the parent of the new commit
        @param blank: Start from a blank state or from a previous commit
        @param force: Start from a blank state or from a previous commit
        @param job_slug: Job that this commit related to
        @return: Dict containing commit data
        """
        data = {
            "force": force,
            "blank": blank,
            "job_slug": job_slug,
            "parent_sha1": parent_sha1,
            "source": source,
            "message": message
        }
        response = self._proxy.call_api(
            route=urljoin(self._route, "commits"),
            http_method=HTTP.POST,
            payload=data
        )

        return response.attributes["sha1"]

    def _end_commit(self, commit_sha1, tag_images=False, with_commit_compare=False):
        """
        End the commit after uploading/deleting files from the dataset.
        @commit_sha1: the commit sha1 to end
        @param tag_images: Will cause images in this commit to be tagged so they can be
            displayed in specific places on front
        @param with_commit_compare: the server will delete the commit if it is identical to the previous
            one (in the case nothing has actually been updated to the server): send last_commit to the server to compare
             with the current commit. (we don't want to create empty commits)
        @return: Dict containing commit data
        """
        data = {'tag_images': tag_images, 'workflow_slug': self.slug}
        if with_commit_compare:
            data["base_commit_sha1"] = self.last_commit
        response = self._proxy.call_api(
            route=urljoin(self._route, "commits", commit_sha1, "end_commit"),
            http_method=HTTP.PUT,
            payload=data
        )
        # Response will be empty if the commit has been deleted
        # (in the case nothing has actually been updated to the server).
        return response.attributes["sha1"] if (response and response.attributes) else None

    def reload(self):
        """
        Performs hard reload for the module attributes and config
        @return: None
        """
        super().reload()
        self._config = Config()
        self._local_commit = None
        # check ownership of local config, unless remove the config
        if not self._validate_config_ownership():
            self._config.local_config = {}

    def put_files(
        self,
        paths,
        message="",
        job_slug=None,
        blank=False,
        override=False,
        force=False,
        upload=False,
        tag_images=False,
        progress_bar_enabled=False,
        git_diff=False
    ):
        """
        Uploads the files and folders given.
        If a folder is given all the relevant files in that folder (that confirms to the regex) will be uploaded.
        @param paths: List of paths or unix-style wildcards
        @param message: String defining the commit message
        @param job_slug: Slug of a job to upload the files to
        @param blank: Start from a blank state or from a previous commit
        @param override: Boolean stating whether or not we should re-upload even if the file already exists
        @param force: Boolean stating whether or not we the new commit should copy files from parent
        @param upload: Boolean gives info if the put files comes from upload or not
        @param tag_images: Boolean indicating if we want to allow only images to be uploaded. used by exp.log_image
        @param progress_bar_enabled: Boolean indicating whenever or not to print a progress bar. In use of the cli
        @param git_diff: upload files from git diff output in addition to the given paths
        @return: string - The Commit sha1 that was created
        """

        last_commit = self.local_commit
        if job_slug is not None:
            # Ugly patch, decided on 12/07/2022 by Elad
            # For experiments, parent commit should be last commit and not start commit
            if type(self).__name__ == 'Experiment':
                last_commit = self.last_commit
            elif last_commit is None and self.start_commit is not None:
                last_commit = self.start_commit["sha1"]

        # Set relevant variables
        filters = ['image'] if tag_images else []
        paths_to_upload = paths if isinstance(paths, list) else [paths]
        if git_diff:
            paths_to_upload += self._handle_git_files()

        # Parses the cnvrgignore and returns a gitignore-like filter object
        ignore_spec = build_cnvrgignore_spec(self._config.root)
        # Get the temp folder path
        temp_folder_path = "{}/{}".format(self.working_dir, DataOwner.TMP_FOLDER_NAME)

        metadata, path_generator = metadata_generator(
            paths=paths_to_upload,
            mime_filters=filters,
            ignore_spec=ignore_spec,
            temp_folder=temp_folder_path
        )

        if metadata["total_files"] == 0:
            return

        # Create a new commit
        commit_sha1 = self._start_commit(
            message=message,
            blank=blank,
            parent_sha1=last_commit,
            job_slug=job_slug,
            force=force
        )

        uploader = self._get_file_uploader(
            paths=path_generator,
            commit=commit_sha1,
            metadata=metadata,
            override=override,
            progress_bar_enabled=progress_bar_enabled
        )
        while uploader.in_progress:
            time.sleep(0.1)

        # TODO: Decide on error behaviour later
        if uploader.errors:
            raise uploader.errors[0]

        if upload:
            self.reload()
            deleter = self._get_remote_file_deleter(commit_sha1)
            while deleter.in_progress:
                time.sleep(0.1)

        # Pass with_commit_compare to end_commit: the server will delete the commit if it is identical to the previous
        # One (in the case where nothing has actually been updated to the server)
        commit = self._end_commit(
            commit_sha1=commit_sha1,
            tag_images=tag_images,
            with_commit_compare=not override
        )

        # Commit will be null if the commit has been deleted
        # In the case nothing has actually been updated to the server, return last_commit
        return commit if commit else self.last_commit

    def _get_file_uploader(self, paths, commit, metadata, override, progress_bar_enabled):
        # this is required for mocking. please don't delete the wrapping function
        return FileUploader(
            data_owner=self,
            paths=paths,
            commit=commit,
            metadata=metadata,
            override=override,
            progress_bar_enabled=progress_bar_enabled
        )

    def _handle_git_files(self):
        return list(filter(None, self.get_git_diff()))  # Clean from falsy values

    def get_git_diff(self):
        """
        collects file names from git diff output
        @return: list of file names from git diff command
        """
        command = "git diff --name-only"
        return os.popen(command).read().strip().split("\n")

    def _get_remote_file_deleter(self, commit_sha1):
        # this is required for mocking. please don't delete the wrapping function
        return RemoteFileDeleter(self, commit=commit_sha1)

    def remove_files(self, paths, message="", progress_bar_enabled=False):
        """
        Removes the given files remotely
        @param paths: List of file paths to remove (regex and wildcards allowed)
        @param message: Commit message
        @param progress_bar_enabled: Boolean indicating whenever or not to print a progress bar. In use of the cli
        @return: Number of files removed
        """
        progress_bar = None
        commit_sha1 = self._start_commit(message=message)

        data = {
            "filter": json.dumps({
                "operator": 'OR',
                "conditions": list_to_multiple_conditions("fullpath", Operators.LIKE, paths),
            })
        }

        meta = self._remove_api_call(commit_sha1=commit_sha1, data=data)

        total_files_count = meta["total"]
        total_files_size = float(meta["total_size"])
        removed_files_count = meta["removed_files_count"]

        if progress_bar_enabled and total_files_size > 0:
            progress_bar = init_progress_bar_for_cli("Removing", total_files_size) if progress_bar_enabled else None

        while True:
            if progress_bar:
                converted_bytes, unit = convert_bytes(float(meta["removed_files_size"]), progress_bar.unit)
                progress_bar.throttled_next(converted_bytes)

            if removed_files_count >= total_files_count:
                break

            meta = self._remove_api_call(commit_sha1=commit_sha1, data=data)
            removed_files_count += meta["removed_files_count"]

        self._end_commit(commit_sha1=commit_sha1)
        return removed_files_count

    def _remove_api_call(self, commit_sha1, data):
        """
        Performs API call to the server with the data to remove
        @param data: Hash containing the paths to remove
        @return: Hash with metadata regarding the remove action
        """
        route = urljoin(self._route, COMMIT_REMOVE_FILES.format(commit_sha1))
        return self._proxy.call_api(
            route=route,
            http_method=HTTP.POST,
            payload=data
        ).meta

    def clone(self, progress_bar_enabled=False, override=False, commit=None, use_cached=False, threads=40,
              query_slug=None):
        """
        Clones the remote project / dataset
        @param progress_bar_enabled: Boolean indicating whenever or not to print a progress bar. In use of the cli
        @param override: Boolean stating whether we should re-clone even if the project / dataset already cloned
        @param commit: [String] sha1 of the commit to clone
        @param use_cached: Boolean stating whether to use nfs cache link or not
        @param threads: number of threads that will be used in order to clone the project
        @return: None
        """
        self.reload()

        if not os.path.exists(self.slug):
            os.makedirs(self.slug)
        elif os.path.exists(self.slug + '/' + CONFIG_FOLDER_NAME):
            if not override:
                # If already cloned and override set to false - won't clone again
                return
            else:
                # If already cloned and override set to true - remove config file so that verify will fail until the
                # data owner is cloned again
                shutil.rmtree(self.slug + '/' + CONFIG_FOLDER_NAME)

        if commit and commit.lower() != 'latest':
            # self.last_commit should be a value returned from the server,
            # and no value should be manually assigned to it.
            # this is a workaround for cloning a specific commit:
            # because FileDownloader uses self.last_commit instead of receiving commit as an arg
            # todo : pass commit to FileDownloader

            self.last_commit = commit

        old_working_dir = self.working_dir
        self.working_dir = "{}/{}".format(old_working_dir, self.slug)

        # Will download files from self.last_commit.
        # In the case of cloning a specific commit, last_commit is actually that specific commit assigned previously

        downloader = FileDownloader(self, progress_bar_enabled=progress_bar_enabled, use_cached=use_cached,
                                    num_workers=threads, query_slug=query_slug)

        while downloader.in_progress:
            time.sleep(1)

        if downloader.errors:
            if os.path.exists(self.slug) and not override:
                os.rmdir(self.slug)
            raise CnvrgError(downloader.errors.args)

        # In the case of cloning a specific commit, last_commit is that specific commit
        # and not the actual data_owner last_commit
        self.local_commit = self.last_commit
        self.save_config(local_config_path=self.working_dir + '/' + CONFIG_FOLDER_NAME)

        # Reload will reset self.last_commit to the actual data_owner last_commit in case it has been changed previously
        self.reload()
        self.working_dir = old_working_dir

        if not cnvrgignore_exists(self.slug):
            create_cnvrgignore(self.slug)

    def upload(self, sync=False, job_slug=None, progress_bar_enabled=False, git_diff=False, message="",
               output_dir=None):
        """
        Uploads files to remote project / dataset
        @param sync: Boolean gives info if the put files comes from sync or not
        @param job_slug: Slug of a job to upload the files to
        @param progress_bar_enabled: Boolean indicating whenever or not to print a progress bar. In use of the cli
        @param git_diff: upload only files from git diff output
        @param message: String defining the commit message
        @param output_dir: String will only upload the files in the output_dir specified
        @return: None
        """

        # We want to make sure we are in a project/dataset folder in order to continue with the action
        self.reload()
        if not self._validate_config_ownership():
            raise CnvrgFileError(error_messages.CONFIG_YAML_NOT_FOUND)

        working_dir = "{0}/{1}".format(self._config.root, output_dir) if output_dir else self._config.root
        old_working_dir = os.getcwd()
        os.chdir(self._config.root)

        new_commit_sha1 = self.put_files(
            [working_dir],
            upload=True,
            job_slug=job_slug,
            progress_bar_enabled=progress_bar_enabled,
            git_diff=git_diff,
            message=message
        )

        if sync:
            self.move_files_from_tmp()

        self.local_commit = new_commit_sha1
        self.save_config()
        os.chdir(old_working_dir)

    def download(self, sync=False, progress_bar_enabled=False, commit_sha1=None):
        """
        Download files from remote project / dataset
        @param sync: Boolean gives info if the put files comes from sync or not
        @param progress_bar_enabled: Boolean indicating whenever or not to print a progress bar. In use of the cli
        @param commit_sha1: Sha1 of a specific commit to download
        @return: None
        """
        # We want to make sure we are in a project/dataset folder in order to continue with the action
        self.reload()
        if not self._validate_config_ownership():
            raise CnvrgFileError(error_messages.CONFIG_YAML_NOT_FOUND)

        context_working_dir = os.getcwd()
        os.chdir(self._config.root)

        """
        download files from last commit on the server and put them in .tmp folder
        will be fetched at the end out of .tmp
        """
        # create .tmp dir
        if not os.path.exists(DataOwner.TMP_FOLDER_NAME):
            os.makedirs(DataOwner.TMP_FOLDER_NAME)

        old_working_dir = self.working_dir
        self.working_dir = "{}/{}".format(self.working_dir, DataOwner.TMP_FOLDER_NAME)
        downloader = ArtifactsDownloader(
            self,
            base_commit_sha1=self.local_commit,
            commit_sha1=commit_sha1 or self.last_commit,
            progress_bar_enabled=progress_bar_enabled
        )
        while downloader.in_progress:
            time.sleep(1)

        if downloader.errors:
            raise CnvrgError(downloader.errors.args)

        self.working_dir = old_working_dir

        # put all files that need to be deleted at .tmp with .deleted extension
        file_deleter = LocalFileDeleter(self)
        while file_deleter.in_progress:
            time.sleep(1)

        # remove all folders needed that should be deleted
        folder_deleter = LocalFileDeleter(self, mode="folders")
        while folder_deleter.in_progress:
            time.sleep(1)

        if not sync:
            self.move_files_from_tmp()
            self.local_commit = commit_sha1 or self.last_commit
            self.save_config()

        os.chdir(context_working_dir)

    def sync(self, job_slug=None, git_diff=False, progress_bar_enabled=False, message='', output_dir=None):
        """
        Sync local project / dataset to remote
        @param job_slug: Slug of a job to upload the files to
        @param git_diff: upload only files from git diff output
        @param progress_bar_enabled: Boolean indicating whenever or not to print a progress bar. In use of the cli
        @param message: String defining the commit message
        @param output_dir: String will only sync the files in the output_dir specified
        @return: None
        """
        try:
            self.download(sync=True, progress_bar_enabled=progress_bar_enabled)
            self.upload(sync=True, progress_bar_enabled=progress_bar_enabled, job_slug=job_slug, git_diff=git_diff,
                        message=message, output_dir=output_dir)
        finally:
            # make sure the tmp folder used by download and upload will be deleted
            if os.path.exists(DataOwner.TMP_FOLDER_NAME):
                shutil.rmtree(DataOwner.TMP_FOLDER_NAME)

    def list_files(self, commit_sha1=None, query=None, query_raw=None, sort="-id"):
        """
        List all files in a specific query
        @param commit_sha1: Sha1 of the commit to list
        @param query: Query slug to list files from
        @param query_raw: Raw query to list files (e.g. {color: yellow})
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @raise: HttpError
        @return: Generator that yields file objects
        """
        return self._list("files", commit_sha1=commit_sha1, query=query, query_raw=query_raw, sort=sort)

    def list_folders(self, commit_sha1=None, query=None, query_raw=None, sort="-id"):
        """
        List all folders in a specific query
        @param commit_sha1: Sha1 of the commit to list
        @param query: Query slug to list folders from
        @param query_raw: Raw query to list folders (e.g. {color: yellow})
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @raise: HttpError
        @return: Generator that yields folder objects
        """
        return self._list("folders", commit_sha1=commit_sha1, query=query, query_raw=query_raw, sort=sort)

    def list(self, commit_sha1=None, query=None, query_raw=None, sort="-id"):
        """
        List all files and folders in a specific query
        @param commit_sha1: Sha1 of the commit to list
        @param query: Query slug to list files and folders from
        @param query_raw: Raw query to list files and folders (e.g. {color: yellow})
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @raise: HttpError
        @return: Generator that yields files and folders objects
        """
        list_folders = self.list_folders(commit_sha1=commit_sha1, query=query, query_raw=query_raw, sort=sort)
        list_files = self.list_files(commit_sha1=commit_sha1, query=query, query_raw=query_raw, sort=sort)
        return chain(list_folders, list_files)

    def move_files_from_tmp(self):
        """
        Takes files from .tmp folder and complete the sync/download/upload operation
        @return: None
        """
        tmp_folder = "{}/{}".format(self.working_dir, DataOwner.TMP_FOLDER_NAME)
        try:
            metadata, path_generator = metadata_generator(['.'])

            if metadata["total_files"] > 0:
                compare = FileCompare(self, paths=path_generator, metadata=metadata)
                while compare.in_progress:
                    time.sleep(1)

            downloaded_files = get_files_and_dirs_recursive(root_dir=tmp_folder, force=True)
            for df in downloaded_files:
                if df == ".tmp/" or df == ".tmp/.cnvrgignore":
                    continue

                if df.endswith('.deleted') or df.endswith('.deleted/'):
                    original_file = re.sub(DataOwner.TMP_FOLDER_NAME, self.working_dir, df)
                    original_file = re.sub(".deleted", "", original_file)
                    if(os.path.isfile(original_file)):
                        os.remove(original_file)
                    elif(os.path.isdir(original_file)):
                        shutil.rmtree(original_file)
                    continue

                new_path = re.sub(DataOwner.TMP_FOLDER_NAME, self.working_dir, df)
                create_dir_if_not_exists(new_path)

                # Prevents moving folders with content. Generally, creation of folders is done when dealing with files
                # inside the folders, except for empty folders. This makes sure that only empty folders will be copied
                if os.path.isdir(df) and (len(os.listdir(df)) != 0 or len(os.listdir(new_path)) != 0):
                    continue

                os.rename(df, new_path)
        finally:
            if os.path.exists(tmp_folder):
                shutil.rmtree(tmp_folder)

    def _get_git_diff_files(self):
        return list(filter(None, self.get_git_diff()))   # Clean from falsy values

    def _list(self, mode, commit_sha1=None, query=None, query_raw=None, sort="-id"):

        if query and query_raw:
            raise CnvrgArgumentsError(error_messages.QUERY_LIST_FAULTY_PARAMS.format("query and query_raw"))

        if query and commit_sha1:
            raise CnvrgArgumentsError(error_messages.QUERY_LIST_FAULTY_PARAMS.format("query and commit_sha1"))

        if not commit_sha1:
            self.reload()
            commit_sha1 = self.local_commit or self.last_commit

        list_object = Folder if mode == "folders" else File

        return api_list_generator(
            context=self._context,
            route=urljoin(self._route, "commits", commit_sha1, "files?mode={}".format(mode)),
            object=list_object,
            sort=sort,
            identifier="fullpath",
            pagination_type="offset",
            data={
                "query": query,
                "query_raw": query_raw
            }
        )



import click

from cnvrgv2.cli.utils import messages
from cnvrgv2.cli.utils.decorators import prepare_command


@click.group(name='experiment')
def experiment_group():
    pass


@experiment_group.command()
@click.option('-t', '--title', default=None, help=messages.EXPERIMENT_HELP_TITLE)
@click.option('-tm', '--templates', default=None, help=messages.EXPERIMENT_HELP_TEMPLATES)
@click.option('-l/-nl', '--local/--no-local', default=False, help=messages.EXPERIMENT_HELP_LOCAL)
@click.option('-c', '--command', prompt=messages.EXPERIMENT_PROMPT_COMMAND, help=messages.EXPERIMENT_HELP_COMMAND)
@click.option('-d', '--datasets', default=None, help=messages.EXPERIMENT_HELP_DATASETS)
@click.option('-v', '--volume', default=None, help=messages.EXPERIMENT_HELP_VOLUME)
@click.option('-sb/-nsb', '--sync-before/--no-sync-before', default=True, help=messages.EXPERIMENT_HELP_SYNC_BEFORE)
@click.option('-sa/-nsa', '--sync-after/--no-sync-after', default=True, help=messages.EXPERIMENT_HELP_SYNC_AFTER)
@click.option('-i', '--image', default=None, help=messages.EXPERIMENT_HELP_IMAGE)
@click.option('-gb', '--git-branch', default=None, help=messages.EXPERIMENT_HELP_GIT_BRANCH)
@click.option('-gc', '--git-commit', default=None, help=messages.EXPERIMENT_HELP_GIT_COMMIT)
@prepare_command()
def run(
    cnvrg,
    logger,
    project,
    title,
    templates,
    local,
    command,
    datasets,
    volume,
    sync_before,
    sync_after,
    image,
    git_branch,
    git_commit
):
    """
      run an experiment
    """
    dataset_objects = None
    volume_object = None
    kwargs = {}
    templates_list = templates.split(",") if templates else None

    if datasets:
        dataset_names = datasets.split(",")
        dataset_objects = [cnvrg.datasets.get(ds_name) for ds_name in dataset_names]

    if volume:
        volume_object = project.volumes.get(volume)

    if image:
        image_name, image_tag = image.split(":")
        kwargs["image"] = cnvrg.images.get(name=image_name, tag=image_tag)

    if git_branch:
        kwargs["git_branch"] = git_branch

    if git_commit:
        kwargs["git_commit"] = git_commit

    experiment = project.experiments.create(
        title=title,
        templates=templates_list,
        local=local,
        command=command,
        datasets=dataset_objects,
        volume=volume_object,
        sync_before=sync_before,
        sync_after=sync_after,
        **kwargs
    )
    success_message = messages.EXPERIMENT_CREATE_SUCCESS.format(experiment.title, experiment.full_href)
    logger.log_and_echo(success_message)


@experiment_group.command()
@click.option('-s', '--slug', default=None, help=messages.EXPERIMENT_HELP_TITLE)
@click.option('-c', '--commit', prompt=messages.EXPERIMENT_PROMPT_COMMIT, default='latest', required=False,
              help=messages.EXPERIMENT_HELP_COMMIT)
@prepare_command()
def merge_to_master(logger, project, experiment, slug, commit):
    # TODO: Slug is not necessary as a parameter (it is as an option, since prepare command uses it).
    #  Change prepare command to send only required arguments

    if project.git:
        logger.log_and_echo(messages.EXPERIMENT_GIT_ERROR_MESSAGE, error=True)
        return
    commit_param = None if commit == 'latest' else commit
    if experiment is None:
        logger.log_and_echo(messages.EXPERIMENT_DOES_NOT_EXIST, error=True)
    sha1 = experiment.merge_to_master(commit_param)
    logger.log_and_echo(messages.EXPERIMENT_MERGE_SUCCESS)
    logger.log_and_echo(messages.COMMIT_SHA1_MESSAGE.format(sha1))


@experiment_group.command()
@click.option('-k', '--key', help='')
@click.option('-v', '--value', help='')
@click.option('-s', '--slug', default=None, help=messages.EXPERIMENT_HELP_TITLE)
@prepare_command()
def log_param(experiment, logger, slug, key, value):
    """
      logging a parameter of an experiment
    """
    if experiment is None:
        logger.log_and_echo(messages.EXPERIMENT_DOES_NOT_EXIST, error=True)
    experiment.log_param(key=key, value=value)
    logger.log_and_echo('{0}: {1} was logged'.format(key, value))


@experiment_group.command()
@click.option('-s', '--slug', default=None, help=messages.EXPERIMENT_HELP_TITLE)
@click.option('-f', '--files', prompt=messages.DATASET_PUT_PROMPT_FILES, help=messages.DATASET_PUT_HELP_FILES)
@click.option('-g', '--git-diff', is_flag=True, default=None, help=messages.DATA_UPLOAD_HELP_GIT_DIFF)
@click.option('-d', '--work-dir', required=False,
              help=messages.EXPERIMENT_HELP_COMMIT)
@prepare_command()
def log_artifacts(cnvrg, logger, project, experiment, slug, files, git_diff, work_dir):
    file_paths = files.split(",")
    if experiment is None:
        logger.log_and_echo(messages.EXPERIMENT_DOES_NOT_EXIST, error=True)
    experiment.log_artifacts(paths=file_paths, git_diff=git_diff, work_dir=work_dir)
    logger.log_and_echo(messages.EXPERIMENT_LOG_ARTIFACTS_SUCCESS)


@experiment_group.command()
@click.option('-s', '--slug', default=None, help=messages.EXPERIMENT_HELP_TITLE)
@click.option('-f', '--files', prompt=messages.DATASET_PUT_PROMPT_FILES, help=messages.DATASET_PUT_HELP_FILES)
@prepare_command()
def log_images(cnvrg, logger, project, experiment, slug, files):
    file_paths = files.split(",")
    if experiment is None:
        logger.log_and_echo(messages.EXPERIMENT_DOES_NOT_EXIST, error=True)
    experiment.log_images(file_paths=file_paths)
    logger.log_and_echo(messages.EXPERIMENT_LOG_IMAGES_SUCCESS)


@experiment_group.command()
@click.option('-s', '--slug', default=None, help=messages.EXPERIMENT_HELP_TITLE)
@prepare_command()
def start_tensorboard(experiment, logger, slug):
    """
      start a tensorboard associated  with the current experiment
    """
    if experiment is None:
        logger.log_and_echo(messages.EXPERIMENT_DOES_NOT_EXIST, error=True)
    experiment.start_tensorboard()
    logger.log_and_echo('tensorboard started!')


@experiment_group.command()
@click.option('-s', '--slug', default=None, help=messages.EXPERIMENT_HELP_TITLE)
@prepare_command()
def stop_tensorboard(experiment, logger, slug):
    """
      stops a tensorboard associated  with the current experiment
    """
    if experiment is None:
        logger.log_and_echo(messages.EXPERIMENT_DOES_NOT_EXIST, error=True)
    experiment.stop_tensorboard()
    logger.log_and_echo('tensorboard session stopped')


@experiment_group.command()
@click.option('-s', '--slug', default=None, help=messages.EXPERIMENT_HELP_TITLE)
@click.option('-c', '--commit', prompt=messages.EXPERIMENT_ARTIFACTS_PROMPT_COMMIT, default='latest', required=False,
              help=messages.EXPERIMENT_ARTIFACTS_HELP_COMMIT)
@prepare_command()
def pull_artifacts(experiment, logger, slug, commit):
    """
      pull artifacts
      default is experiment's last commit
    """
    if experiment is None:
        logger.log_and_echo(messages.EXPERIMENT_DOES_NOT_EXIST, error=True)
    commit_param = None if commit == 'latest' else commit
    experiment.pull_artifacts(commit_sha1=commit_param)
    logger.log_and_echo(messages.EXPERIMENT_PULL_ARTIFACTS_SUCCESS)



import subprocess
import time

import click

from cnvrgv2.cli.utils import messages
from cnvrgv2.cli.utils.decorators import prepare_command
from cnvrgv2.config import error_messages


@click.group(name='ssh')
def ssh_group():
    pass


@ssh_group.command()
@click.argument('job_id')
@click.option("-p", "--port", help=messages.SSH_HELP_PORT, default=2222)
@click.option("-u", "--username", help=messages.SSH_HELP_USERNAME, default=None)
@click.option("-pw", "--password", help=messages.SSH_HELP_PASSWORD, default=None)
@click.option("-kc", "--kubeconfig", help=messages.SSH_HELP_KUBECTL, default=None)
@prepare_command()
def start(cnvrg, logger, job_id, port, username, password, kubeconfig):
    """
    Creates an ssh session to the pod of the given JOB_ID
    """
    logger.log_and_echo(messages.SSH_STARTING_SESSION)
    ssh = cnvrg.ssh.create_ssh(job_id)
    ssh.start(username, password)

    while True:
        ssh.reload()
        if ssh.ssh_status == "in_progress":
            click.echo(messages.SSH_WAITING_FOR_READY)
            time.sleep(3)
            continue
        elif ssh.ssh_status == "finished":
            break
        else:
            message = error_messages.SSH_FAILED_TO_START.format(ssh.ssh_status)
            logger.log_and_echo(message, error=True)

    if ssh.pod_name is None or ssh.password is None or ssh.username is None:
        message = error_messages.SSH_FAILED_GET_REQUIRED_PARAMS.format(ssh.ssh_status)
        logger.log_and_echo(message, error=True)

    connection_message = messages.SSH_READY.format(port, ssh.username, ssh.password)
    logger.log_and_echo(connection_message)

    run_portforward_command(ssh.pod_name, port, kubeconfig, ssh.namespace)


def run_portforward_command(pod_name, port, kubeconfig, namespace):
    command = "kubectl"
    if kubeconfig is not None:
        command = "kubectl --kubeconfig={}".format(kubeconfig)

    bashCommand = "{} -n {} port-forward {} {}:22".format(command, namespace, pod_name, port)
    click.echo("\nrunning command {}".format(bashCommand))
    subprocess.check_output(['bash', '-c', bashCommand])



import re

from cnvrgv2.config import error_messages, routes
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.modules.images.image import Image
from cnvrgv2.modules.images.utils import ImageLogos, TAG_VALIDATION_REGEX, NAME_VALIDATION_REGEX
from cnvrgv2.proxy import HTTP, Proxy
from cnvrgv2.utils.api_list_generator import api_list_generator
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.utils.url_utils import urljoin


class ImagesClient:
    def __init__(self, organization):
        self._context = Context(context=organization._context)
        scope = self._context.get_scope(SCOPE.ORGANIZATION)

        self._proxy = Proxy(context=self._context)
        self._route = routes.IMAGES_BASE.format(scope["organization"])

    def list(self, sort="-id"):
        """
        List all images in a specific organization
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @raise: HttpError
        @return: Generator that yields image objects
        """
        return api_list_generator(
            context=self._context,
            route=self._route,
            object=Image,
            sort=sort
        )

    def create(self, name, tag, registry="local", logo="", custom=False, dockerfile="", readme=""):
        """
        Create a new image
        @param name: [String] The image name
        @param tag: [String] The image tag
        @param registry: [String] The registry slug (Must be a custom registry to build a custom image)
        @param logo: [String] The image logo (to be displayed in the UI) - defaults to docker
        @param custom: [String] whether the image is custom (needs to be built)
        @param dockerfile: [String] The dockerfile with which to build the image
        @param readme: [String] Image readme (will be displayed in the UI)
        @return: [Image object] The newly created image
        """

        if not tag or not isinstance(tag, str) or not re.match(TAG_VALIDATION_REGEX, tag):
            raise CnvrgArgumentsError(error_messages.IMAGE_BAD_TAG)

        if not name or not isinstance(name, str) or not re.match(NAME_VALIDATION_REGEX, name):
            raise CnvrgArgumentsError(error_messages.IMAGE_BAD_NAME)

        if not registry or not isinstance(registry, str):
            raise CnvrgArgumentsError(error_messages.IMAGE_BAD_REGISTRY)

        if not ImageLogos.validate_icon(logo):
            raise CnvrgArgumentsError(error_messages.IMAGE_BAD_LOGO)

        if custom and not dockerfile:
            raise CnvrgArgumentsError(error_messages.IMAGE_CUSTOM_DOCKERFILE_REQUIRED)

        attributes = {
            "tag": tag,
            "name": name,
            "logo": logo,
            "is_custom": bool(custom),
            "registry_slug": registry,
            "readme": readme,
            "dockerfile": dockerfile,
        }
        response = self._proxy.call_api(
            route=self._route,
            http_method=HTTP.POST,
            payload=JAF.serialize(type="image", attributes=attributes)
        )

        slug = response.attributes['slug']
        return Image(context=self._context, slug=slug)

    def get(self, slug=None, name=None, tag=None):
        """
        Retrieves an Image by the given slug, or name and tag
        @param slug: [String] The slug of the requested image
        @param tag: [String] The image tag
        @param name: [String] The image name
        @return: Image object
        """

        if slug and isinstance(slug, str):
            return Image(context=self._context, slug=slug)
        elif not slug and (name and tag):
            res_attributes = self._proxy.call_api(
                route=urljoin(self._route, routes.GET_BY_NAME_SUFFIX),
                http_method=HTTP.GET,
                payload={"image_name": name, "image_tag": tag}
            ).attributes

            return Image(context=self._context, slug=res_attributes['slug'], attributes=res_attributes)
        else:
            raise CnvrgArgumentsError(error_messages.IMAGE_GET_FAULTY_PARAMS)

    def delete(self, slug):
        """
        Delete an image by its slug
        @param slug: [String] Image slug
        @return: None
        """

        Image(context=self._context, slug=slug).delete()



from cnvrgv2.config.error_messages import NOT_A_VOLUME_OBJECT, NOT_A_DATASET_LIST_OBJECT
from cnvrgv2.context import SCOPE
from cnvrgv2.config import routes
from cnvrgv2.errors import CnvrgArgumentsError

from cnvrgv2.modules.base.workflows_base import WorkflowsBase
from cnvrgv2.modules.dataset import Dataset
from cnvrgv2.modules.volumes.volume import Volume
from cnvrgv2.modules.workflows import Workspace, NotebookType
from cnvrgv2.utils.validators import validate_types_in_list


class WorkspacesClient(WorkflowsBase):
    def __init__(self, project):
        super().__init__(Workspace, "NotebookSession", project._context)

        scope = self._context.get_scope(SCOPE.PROJECT)
        self._route = routes.WORKSPACES_BASE.format(scope["organization"], scope["project"], "workspaces")

    def create(
            self,
            title=None,
            templates=None,
            notebook_type=NotebookType.JUPYTER_LAB,
            volume=None,
            datasets=None,
            *args,
            **kwargs):
        """
        Creates a new workspace with the given name
        @param title: Name of the workflow
        @param templates: List of template names to be used
        @param notebook_type: The type of the created notebook. Use NotebookType enum
        @param volume: A volume to attach to this workspace. (type: Volume object).
        @param datasets: List of datasets to connect with the workspace.
        @param kwargs: rest of optional attributes for creation
            image: Image object to create workspace with
            queue: Name of the queue to run this job on
        TODO: Add a list of optional attributes
        @return: The newly created workflow object
        """
        kwargs = {
            "notebook_type": notebook_type,
            **kwargs
        }

        if datasets and not validate_types_in_list(datasets, Dataset):
            raise CnvrgArgumentsError(NOT_A_DATASET_LIST_OBJECT)
        elif datasets:
            kwargs["job_datasets"] = [ds.as_request_params() for ds in datasets]

        if volume and not isinstance(volume, Volume):
            raise CnvrgArgumentsError(NOT_A_VOLUME_OBJECT)
        elif volume:
            kwargs["external_disk_slug"] = volume.slug

        return super().create(title, templates, *args, **kwargs)



import os

import click
from prettytable import PrettyTable

from cnvrgv2.cli.utils import messages
from cnvrgv2.cli.utils.decorators import prepare_command
from cnvrgv2.config import CONFIG_FOLDER_NAME, Config
from cnvrgv2.config import error_messages
from cnvrgv2.utils.storage_utils import cnvrgignore_exists, create_cnvrgignore


@click.group(name='project')
def project_group():
    pass


@project_group.command()
@click.option('-n', '--name', prompt=messages.PROJECT_PROMPT_CREATE, help=messages.PROJECT_HELP_CREATE)
@prepare_command()
def create(cnvrg, logger, name):
    """
    Creates a new project, and associates the current folder with the created project.
    It is recommended to create a new project in an empty folder.
    To create a project from a folder that contains content please refer to project link command.
    """
    if os.listdir():
        click.confirm(messages.PROJECT_CREATE_FOLDER_NOT_EMPTY, default=False, abort=True)

    logger.log_and_echo(messages.PROJECT_CREATING_PROJECT.format(name))
    new_project = cnvrg.projects.create(name)

    new_project.local_commit = new_project.commit

    logger.log_and_echo(messages.PROJECT_CONFIGURING_FOLDER)
    new_project.save_config()

    logger.log_and_echo(messages.PROJECT_CREATE_SUCCESS.format(new_project.title))


@project_group.command()
@click.option('-n', '--name', prompt=messages.PROJECT_PROMPT_CLONE, help=messages.PROJECT_PROMPT_CLONE)
@click.option('-o', '--override', is_flag=True, default=False, help=messages.PROJECT_HELP_CLONE_OVERRIDE)
@click.option('-c', '--commit', prompt=messages.DATA_PROMPT_COMMIT, default="latest",
              help=messages.PROJECT_HELP_CLONE_COMMIT)
@prepare_command()
def clone(cnvrg, logger, name, override, commit):
    """
    Clones the given project to local folder
    """
    project = cnvrg.projects.get(name)
    logger.info(messages.LOG_CLONING_PROJECT.format(name))
    if os.path.exists(project.slug + '/' + CONFIG_FOLDER_NAME) and not override:
        logger.log_and_echo(messages.PROJECT_CLONE_SKIP.format(name))
        return
    project.clone(progress_bar_enabled=True, override=override, commit=commit)
    success_message = messages.PROJECT_CLONE_SUCCESS.format(name)
    logger.log_and_echo(success_message)


@project_group.command(name='list')
@prepare_command()
def list_command(cnvrg, logger):
    """
    list projects title
    """
    projects = cnvrg.projects.list()
    domain = cnvrg._context.domain
    organization = cnvrg._context.organization
    table = PrettyTable()
    table.field_names = ["Project Title", "Project Link", "Public"]
    table.align["Project Title"] = "l"
    table.align["Project Link"] = "l"
    table.align["Public"] = "l"
    for project in projects:
        url = "{0}/{1}/projects/{2}".format(domain, organization, project.slug)
        table.add_row([project.title, url, project.public])
    click.echo(table)


@project_group.command()
@click.option('-n', '--name', prompt=messages.PROJECT_PROMPT_DELETE, help=messages.PROJECT_HELP_DELETE)
@prepare_command()
def delete(cnvrg, logger, name):
    """
    Deletes the current project
    """
    click.confirm(messages.PROJECT_DELETE_CONFIRM.format(name), default=False, abort=True)
    cnvrg.projects.delete([name])
    success_message = messages.PROJECT_DELETE_SUCCESS.format(name)
    logger.log_and_echo(success_message)


@project_group.command()
@click.option('-g', '--git-diff', is_flag=True, help=messages.DATA_UPLOAD_HELP_GIT_DIFF)
@prepare_command()
def upload(project, logger, git_diff):
    """
    Uploads updated files from the current project folder
    """
    project.upload(progress_bar_enabled=True, git_diff=git_diff)
    logger.log_and_echo(messages.DATA_UPLOAD_SUCCESS)


@project_group.command()
@click.option('-c', '--commit', default=None, help=messages.PROJECT_DOWNLOAD_HELP_COMMIT)
@prepare_command()
def download(project, logger, commit):
    """
    Downloads updated files to the current project folder
    """
    project.download(progress_bar_enabled=True, commit_sha1=commit)
    logger.log_and_echo(messages.DATA_DOWNLOAD_SUCCESS)


@project_group.command()
@click.option('-n', '--name', prompt=messages.PROJECT_LINK_PROMPT_NAME, help=messages.PROJECT_LINK_HELP_NAME)
@prepare_command()
def link(cnvrg, logger, name):
    """
    Links the current directory with a new project.
    This command will create a new project and upload the current folder content to the newly created project.
    """
    curr_dir_project_name = Config().data_owner_slug
    if curr_dir_project_name:
        error_message = error_messages.DIRECTORY_ALREADY_LINKED.format(curr_dir_project_name)
        logger.log_and_echo(error_message, error=True)

    logger.log_and_echo(messages.PROJECT_CREATE_NEW.format(name))
    new_project = cnvrg.projects.create(name=name)
    new_project.local_commit = new_project.commit

    logger.log_and_echo(messages.PROJECT_CONFIGURING_FOLDER)
    new_project.save_config()

    logger.log_and_echo(messages.PROJECT_UPLOAD)
    new_project.upload(progress_bar_enabled=True)

    success_message = messages.PROJECT_LINK_SUCCESS.format(new_project.title)
    logger.log_and_echo(success_message)


@project_group.command()
@click.option('-n', '--name', prompt=messages.PROJECT_LINK_GIT_PROMPT_NAME, help=messages.PROJECT_LINK_GIT_HELP_NAME)
@click.option('-s', '--soft', is_flag=True, help=messages.SOFT_LINK_GIT_PROJECT)
@prepare_command()
def link_git(cnvrg, logger, name, soft):
    """
     Links the current directory, which is expected to be a git directory, with an existing cnvrg project
     that is configured to be a git project.
    """
    git_proj = cnvrg.projects.get(name)
    if not git_proj:
        project_not_exist_error = error_messages.PROJECT_NOT_EXIST.format(name)
        logger.log_and_echo(project_not_exist_error, error=True)

    if not git_proj.git:
        error_message = error_messages.NOT_A_GIT_PROJECT.format(git_proj.title)
        logger.log_and_echo(error_message, error=True)

    if soft and os.path.exists(CONFIG_FOLDER_NAME):
        logger.log_and_echo(messages.PROJECT_GIT_LINK_SKIP.format(git_proj.title))
        return

    logger.log_and_echo(messages.PROJECT_CONFIGURING_FOLDER)
    git_proj.local_commit = git_proj.commit
    git_proj.save_config()

    if not cnvrgignore_exists(os.getcwd()):
        create_cnvrgignore(os.getcwd())

    success_message = messages.PROJECT_LINK_SUCCESS.format(git_proj.title)
    logger.log_and_echo(success_message)


@project_group.command()
@click.option('-j', '--job-slug', default=None, help=messages.PROJECT_JOB_SLUG_HELP_NAME)
@click.option('-g', '--git-diff', is_flag=True, default=None, help=messages.DATA_UPLOAD_HELP_GIT_DIFF)
@click.option('-m', '--message', default="", help=messages.DATA_COMMIT_MESSAGE)
@click.option('-o', '--output-dir', default=None, help=messages.SYNC_OUTPUT_DIR)
@prepare_command()
def sync(project, logger, message, job_slug=None, git_diff=None, output_dir=None):
    """
    Sync local project to remote
    """
    logger.log_and_echo('Syncing...')
    project.sync(job_slug=job_slug, git_diff=git_diff, progress_bar_enabled=True, message=message,
                 output_dir=output_dir)
    logger.log_and_echo(messages.PROJECT_SYNC_SUCCESS.format(project.slug))


@project_group.command()
@click.option('-n', '--name', prompt=messages.PROJECT_PROMPT_NAME, help=messages.PROJECT_HELP_NAME)
@click.option('-f', '--files', prompt=messages.DATASET_PUT_PROMPT_FILES, help=messages.DATASET_PUT_HELP_FILES)
@click.option('-fc', '--force', is_flag=True, help=messages.DATA_UPLOAD_HELP_FORCE)
@click.option('-or', '--override', is_flag=True, help=messages.DATA_UPLOAD_HELP_OVERRIDE)
@click.option('-gd', '--git-diff', is_flag=True, help=messages.DATA_UPLOAD_HELP_GIT_DIFF)
@prepare_command()
def put(cnvrg, logger, name, files, force, override, git_diff):
    """
    Uploads the given files to the given project
    """
    file_paths = files.split(",")
    project = cnvrg.projects.get(name)
    sha1 = project.put_files(
        paths=file_paths,
        progress_bar_enabled=True,
        git_diff=git_diff,
        force=force,
        override=override
    )
    logger.log_and_echo("\n" + messages.FILES_UPLOAD_SUCCESS)
    logger.log_and_echo(messages.COMMIT_SHA1_MESSAGE.format(sha1))


@project_group.command()
@click.option('-n', '--name', prompt=messages.PROJECT_PROMPT_NAME, help=messages.PROJECT_HELP_DELETE)
@click.option('-f', '--files', prompt=messages.DATASET_REMOVE_PROMPT_FILES, help=messages.DATASET_REMOVE_HELP_FILES)
@click.option('-m', '--message', help=messages.DATA_COMMIT_MESSAGE, default="")
@prepare_command()
def rm(cnvrg, logger, name, files, message):
    """
    Removes the given files remotely
    """
    file_paths = files.split(",")
    project = cnvrg.projects.get(name)
    removed_files_count = project.remove_files(paths=file_paths, message=message, progress_bar_enabled=True)
    str_count = str(removed_files_count)
    pluralize = "File{}".format('s'[:removed_files_count ^ 1])  # Pluralizes the word (File/Files)
    logger.log_and_echo("{} {} removed.".format(str_count, pluralize))
